from typing import Any, cast

import torch
import wandb
from torch.utils.data import DataLoader
from tqdm import tqdm

from sae_lens.training.sparse_autoencoder import SparseAutoencoder


def train_toy_sae(
    sparse_autoencoder: SparseAutoencoder,
    activation_store: torch.Tensor,  # TODO: this type seems strange / wrong
    batch_size: int = 1024,
    feature_sampling_window: int = 100,  # how many training steps between resampling the features / considiring neurons dead
    dead_feature_window: int = 2000,  # how many training steps before a feature is considered dead
    dead_feature_threshold: float = 1e-8,  # how infrequently a feature has to be active to be considered dead
    use_wandb: bool = False,
    wandb_log_frequency: int = 50,
):
    """
    Takes an SAE and a bunch of activations and does a bunch of training steps
    """

    # TODO: this type seems strange
    dataloader = iter(
        DataLoader(cast(Any, activation_store), batch_size=batch_size, shuffle=True)
    )
    optimizer = torch.optim.Adam(sparse_autoencoder.parameters())
    sparse_autoencoder.train()
    frac_active_list = []  # track active features

    n_training_steps = 0
    n_training_tokens = 0

    pbar = tqdm(dataloader, desc="Training SAE")
    for _, batch in enumerate(pbar):
        batch = next(dataloader)
        # Make sure the W_dec is still zero-norm
        if sparse_autoencoder.normalize_sae_decoder:
            sparse_autoencoder.set_decoder_norm_to_unit_norm()

        # Forward and Backward Passes
        optimizer.zero_grad()
        sae_out, feature_acts, loss, mse_loss, l1_loss, _ = sparse_autoencoder(batch)
        loss.backward()
        if sparse_autoencoder.normalize_sae_decoder:
            sparse_autoencoder.remove_gradient_parallel_to_decoder_directions()
        optimizer.step()

        n_training_tokens += batch_size

        with torch.no_grad():
            # Calculate the sparsities, and add it to a list
            act_freq_scores = (feature_acts.abs() > 0).float().sum(0)
            frac_active_list.append(act_freq_scores)

            if len(frac_active_list) > feature_sampling_window:
                frac_active_in_window = torch.stack(
                    frac_active_list[-feature_sampling_window:], dim=0
                )
                feature_sparsity = frac_active_in_window.sum(0) / (
                    feature_sampling_window * batch_size
                )
            else:
                # use the whole list
                frac_active_in_window = torch.stack(frac_active_list, dim=0)
                feature_sparsity = frac_active_in_window.sum(0) / (
                    len(frac_active_list) * batch_size
                )

            l0 = (feature_acts > 0).float().sum(-1).mean()
            l2_norm = torch.norm(feature_acts, dim=1).mean()

            l2_norm_in = torch.norm(batch, dim=-1)
            l2_norm_out = torch.norm(sae_out, dim=-1)
            l2_norm_ratio = l2_norm_out / (1e-6 + l2_norm_in)

            if use_wandb and ((n_training_steps + 1) % wandb_log_frequency == 0):
                wandb.log(
                    {
                        "details/n_training_tokens": n_training_tokens,
                        "losses/mse_loss": mse_loss.item(),
                        "losses/l1_loss": l1_loss.item(),
                        "losses/overall_loss": loss.item(),
                        "metrics/l0": l0.item(),
                        "metrics/l2": l2_norm.item(),
                        "metrics/l2_ratio": l2_norm_ratio.mean().item(),
                        "sparsity/below_1e-5": (feature_sparsity < 1e-5)
                        .float()
                        .mean()
                        .item(),
                        "sparsity/below_1e-6": (feature_sparsity < 1e-6)
                        .float()
                        .mean()
                        .item(),
                        "sparsity/n_dead_features": (
                            feature_sparsity < dead_feature_threshold
                        )
                        .float()
                        .mean()
                        .item(),
                    },
                    step=n_training_steps,
                )

                if (n_training_steps + 1) % (wandb_log_frequency * 100) == 0:
                    log_feature_sparsity = torch.log10(feature_sparsity + 1e-8)
                    wandb.log(
                        {
                            "plots/feature_density_histogram": wandb.Histogram(
                                log_feature_sparsity.tolist()
                            ),
                        },
                        step=n_training_steps,
                    )

            pbar.set_description(
                f"{n_training_steps}| MSE Loss {mse_loss.item():.3f} | L0 {l0.item():.3f}"
            )
            pbar.update(batch_size)

        # If we did checkpointing we'd do it here.

        n_training_steps += 1

    return sparse_autoencoder



import json
from typing import Protocol

import torch
from huggingface_hub import hf_hub_download
from safetensors import safe_open

from sae_lens import __version__
from sae_lens.training.config import LanguageModelSAERunnerConfig


# loaders take in a repo_id, folder_name, device, and whether to force download, and returns a tuple of config and state_dict
class PretrainedSaeLoader(Protocol):
    def __call__(
        self,
        repo_id: str,
        folder_name: str,
        device: str | torch.device | None = None,
        force_download: bool = False,
    ) -> tuple[LanguageModelSAERunnerConfig, dict[str, torch.Tensor]]: ...


def sae_lens_loader(
    repo_id: str,
    folder_name: str,
    device: str | torch.device | None = None,
    force_download: bool = False,
) -> tuple[LanguageModelSAERunnerConfig, dict[str, torch.Tensor]]:
    cfg_filename = f"{folder_name}/cfg.json"
    cfg_path = hf_hub_download(
        repo_id=repo_id, filename=cfg_filename, force_download=force_download
    )

    weights_filename = f"{folder_name}/sae_weights.safetensors"
    sae_path = hf_hub_download(
        repo_id=repo_id, filename=weights_filename, force_download=force_download
    )

    return load_pretrained_sae_lens_sae_components(cfg_path, sae_path, device)


def load_pretrained_sae_lens_sae_components(
    cfg_path: str, weight_path: str, device: str | torch.device | None = None
) -> tuple[LanguageModelSAERunnerConfig, dict[str, torch.Tensor]]:
    with open(cfg_path, "r") as f:
        config = json.load(f)
    var_names = LanguageModelSAERunnerConfig.__init__.__code__.co_varnames
    # filter config for varnames
    config = {k: v for k, v in config.items() if k in var_names}
    config["verbose"] = False
    config["device"] = device

    # TODO: if we change our SAE implementation such that old versions need conversion to be
    # loaded, we can inspect the original "sae_lens_version" and apply a conversion function here.
    config["sae_lens_version"] = __version__

    config = LanguageModelSAERunnerConfig(**config)

    tensors = {}
    with safe_open(weight_path, framework="pt", device=device) as f:  # type: ignore
        for k in f.keys():
            tensors[k] = f.get_tensor(k)

    # old saves may not have scaling factors.
    if "scaling_factor" not in tensors:
        assert isinstance(config.d_sae, int)
        tensors["scaling_factor"] = torch.ones(
            config.d_sae, dtype=config.dtype, device=config.device
        )

    return config, tensors


# TODO: add more loaders for other SAEs not trained by us

NAMED_PRETRAINED_SAE_LOADERS: dict[str, PretrainedSaeLoader] = {
    "sae_lens": sae_lens_loader,
}



from types import SimpleNamespace
from typing import Optional

import torch
import tqdm


def weighted_average(points: torch.Tensor, weights: torch.Tensor):
    weights = weights / weights.sum()
    return (points * weights.view(-1, 1)).sum(dim=0)


@torch.no_grad()
def geometric_median_objective(
    median: torch.Tensor, points: torch.Tensor, weights: torch.Tensor
) -> torch.Tensor:

    norms = torch.linalg.norm(points - median.view(1, -1), dim=1)

    return (norms * weights).sum()


def compute_geometric_median(
    points: torch.Tensor,
    weights: Optional[torch.Tensor] = None,
    eps: float = 1e-6,
    maxiter: int = 100,
    ftol: float = 1e-20,
    do_log: bool = False,
):
    """
    :param points: ``torch.Tensor`` of shape ``(n, d)``
    :param weights: Optional ``torch.Tensor`` of shape :math:``(n,)``.
    :param eps: Smallest allowed value of denominator, to avoid divide by zero.
        Equivalently, this is a smoothing parameter. Default 1e-6.
    :param maxiter: Maximum number of Weiszfeld iterations. Default 100
    :param ftol: If objective value does not improve by at least this `ftol` fraction, terminate the algorithm. Default 1e-20.
    :param do_log: If true will return a log of function values encountered through the course of the algorithm
    :return: SimpleNamespace object with fields
        - `median`: estimate of the geometric median, which is a ``torch.Tensor`` object of shape :math:``(d,)``
        - `termination`: string explaining how the algorithm terminated.
        - `logs`: function values encountered through the course of the algorithm in a list (None if do_log is false).
    """
    with torch.no_grad():

        if weights is None:
            weights = torch.ones((points.shape[0],), device=points.device)
        # initialize median estimate at mean
        new_weights = weights
        median = weighted_average(points, weights)
        objective_value = geometric_median_objective(median, points, weights)
        if do_log:
            logs = [objective_value]
        else:
            logs = None

        # Weiszfeld iterations
        early_termination = False
        pbar = tqdm.tqdm(range(maxiter))
        for _ in pbar:
            prev_obj_value = objective_value

            norms = torch.linalg.norm(points - median.view(1, -1), dim=1)
            new_weights = weights / torch.clamp(norms, min=eps)
            median = weighted_average(points, new_weights)
            objective_value = geometric_median_objective(median, points, weights)

            if logs is not None:
                logs.append(objective_value)
            if abs(prev_obj_value - objective_value) <= ftol * objective_value:
                early_termination = True
                break

            pbar.set_description(f"Objective value: {objective_value:.4f}")

    median = weighted_average(points, new_weights)  # allow autodiff to track it
    return SimpleNamespace(
        median=median,
        new_weights=new_weights,
        termination=(
            "function value converged within tolerance"
            if early_termination
            else "maximum iterations reached"
        ),
        logs=logs,
    )


if __name__ == "__main__":
    import time

    TOLERANCE = 1e-2

    dim1 = 10000
    dim2 = 768
    device = "cuda" if torch.cuda.is_available() else "cpu"

    sample = (
        torch.randn((dim1, dim2), device=device) * 100
    )  # seems to be the order of magnitude of the actual use case
    weights = torch.randn((dim1,), device=device)

    torch.tensor(weights, device=device)

    tic = time.perf_counter()
    new = compute_geometric_median(sample, weights=weights, maxiter=100)
    print(f"new code takes {time.perf_counter()-tic} seconds!")



import os
from dataclasses import dataclass, field
from typing import Any, Optional, cast

import torch
import wandb

from sae_lens import __version__

DTYPE_MAP = {
    "torch.float32": torch.float32,
    "torch.float64": torch.float64,
    "torch.float16": torch.float16,
    "torch.bfloat16": torch.bfloat16,
}


@dataclass
class LanguageModelSAERunnerConfig:
    """
    Configuration for training a sparse autoencoder on a language model.
    """

    # Data Generating Function (Model + Training Distibuion)
    model_name: str = "gelu-2l"
    model_class_name: str = "HookedTransformer"
    hook_point: str = "blocks.{layer}.hook_mlp_out"
    hook_point_eval: str = "blocks.{layer}.attn.pattern"
    hook_point_layer: int | list[int] = 0
    hook_point_head_index: Optional[int] = None
    dataset_path: str = "NeelNanda/c4-tokenized-2b"
    is_dataset_tokenized: bool = True
    context_size: int = 128
    use_cached_activations: bool = False
    cached_activations_path: Optional[str] = (
        None  # Defaults to "activations/{dataset}/{model}/{full_hook_name}_{hook_point_head_index}"
    )

    # SAE Parameters
    d_in: int = 512
    d_sae: Optional[int] = None
    b_dec_init_method: str = "geometric_median"
    expansion_factor: int | list[int] = 4
    activation_fn: str = "relu"  # relu, tanh-relu
    normalize_sae_decoder: bool = True
    noise_scale: float = 0.0
    from_pretrained_path: Optional[str] = None
    apply_b_dec_to_input: bool = True
    decoder_orthogonal_init: bool = False

    # Activation Store Parameters
    n_batches_in_buffer: int = 20
    training_tokens: int = 2_000_000
    finetuning_tokens: int = 0
    store_batch_size: int = 32
    train_batch_size: int = 4096

    # Misc
    device: str | torch.device = "cpu"
    seed: int = 42
    dtype: str | torch.dtype = "float32"  # type: ignore #
    prepend_bos: bool = True

    # Training Parameters

    ## Batch size
    train_batch_size: int = 4096

    ## Adam
    adam_beta1: float | list[float] = 0
    adam_beta2: float | list[float] = 0.999

    ## Loss Function
    mse_loss_normalization: Optional[str] = None
    l1_coefficient: float | list[float] = 1e-3
    lp_norm: float | list[float] = 1

    ## Learning Rate Schedule
    lr: float | list[float] = 3e-4
    lr_scheduler_name: str | list[str] = (
        "constant"  # constant, cosineannealing, cosineannealingwarmrestarts
    )
    lr_warm_up_steps: int | list[int] = 500
    lr_end: float | list[float] | None = (
        None  # only used for cosine annealing, default is lr / 10
    )
    lr_decay_steps: int | list[int] = 0
    n_restart_cycles: int | list[int] = 1  # used only for cosineannealingwarmrestarts

    ## FineTuning
    finetuning_method: Optional[str] = None  # scale, decoder or unrotated_decoder

    # Resampling protocol args
    use_ghost_grads: bool | list[bool] = (
        False  # want to change this to true on some timeline.
    )
    feature_sampling_window: int = 2000
    dead_feature_window: int = 1000  # unless this window is larger feature sampling,

    dead_feature_threshold: float = 1e-8

    # WANDB
    log_to_wandb: bool = True
    log_activations_store_to_wandb: bool = False
    log_optimizer_state_to_wandb: bool = False
    wandb_project: str = "mats_sae_training_language_model"
    wandb_id: Optional[str] = None
    run_name: Optional[str] = None
    wandb_entity: Optional[str] = None
    wandb_log_frequency: int = 10

    # Misc
    resume: bool = False
    n_checkpoints: int = 0
    checkpoint_path: str = "checkpoints"
    verbose: bool = True
    model_kwargs: dict[str, Any] = field(default_factory=dict)
    sae_lens_version: str = field(default_factory=lambda: __version__)
    sae_lens_training_version: str = field(default_factory=lambda: __version__)

    def __post_init__(self):
        if self.use_cached_activations and self.cached_activations_path is None:
            self.cached_activations_path = _default_cached_activations_path(
                self.dataset_path,
                self.model_name,
                self.hook_point,
                self.hook_point_head_index,
            )

        if not isinstance(self.expansion_factor, list):
            self.d_sae = self.d_in * self.expansion_factor
        self.tokens_per_buffer = (
            self.train_batch_size * self.context_size * self.n_batches_in_buffer
        )

        if self.run_name is None:
            self.run_name = f"{self.d_sae}-L1-{self.l1_coefficient}-LR-{self.lr}-Tokens-{self.training_tokens:3.3e}"

        if self.b_dec_init_method not in ["geometric_median", "mean", "zeros"]:
            raise ValueError(
                f"b_dec_init_method must be geometric_median, mean, or zeros. Got {self.b_dec_init_method}"
            )
        if self.b_dec_init_method == "zeros":
            print(
                "Warning: We are initializing b_dec to zeros. This is probably not what you want."
            )

        if isinstance(self.dtype, str) and self.dtype not in DTYPE_MAP:
            raise ValueError(
                f"dtype must be one of {list(DTYPE_MAP.keys())}. Got {self.dtype}"
            )
        elif isinstance(self.dtype, str):
            self.dtype: torch.dtype = DTYPE_MAP[self.dtype]

        # if we use decoder fine tuning, we can't be applying b_dec to the input
        if (self.finetuning_method == "decoder") and (self.apply_b_dec_to_input):
            raise ValueError(
                "If we are fine tuning the decoder, we can't be applying b_dec to the input.\nSet apply_b_dec_to_input to False."
            )

        self.device: str | torch.device = torch.device(self.device)

        if self.lr_end is None:
            if isinstance(self.lr, list):
                self.lr_end = [lr / 10 for lr in self.lr]
            else:
                self.lr_end = self.lr / 10
        unique_id = self.wandb_id
        if unique_id is None:
            unique_id = cast(
                Any, wandb
            ).util.generate_id()  # not sure why this type is erroring
        self.checkpoint_path = f"{self.checkpoint_path}/{unique_id}"

        if self.verbose:
            print(
                f"Run name: {self.d_sae}-L1-{self.l1_coefficient}-LR-{self.lr}-Tokens-{self.training_tokens:3.3e}"
            )
            # Print out some useful info:
            n_tokens_per_buffer = (
                self.store_batch_size * self.context_size * self.n_batches_in_buffer
            )
            print(f"n_tokens_per_buffer (millions): {n_tokens_per_buffer / 10 ** 6}")
            n_contexts_per_buffer = self.store_batch_size * self.n_batches_in_buffer
            print(
                f"Lower bound: n_contexts_per_buffer (millions): {n_contexts_per_buffer / 10 ** 6}"
            )

            total_training_steps = (
                self.training_tokens + self.finetuning_tokens
            ) // self.train_batch_size
            print(f"Total training steps: {total_training_steps}")

            total_wandb_updates = total_training_steps // self.wandb_log_frequency
            print(f"Total wandb updates: {total_wandb_updates}")

            # how many times will we sample dead neurons?
            # assert self.dead_feature_window <= self.feature_sampling_window, "dead_feature_window must be smaller than feature_sampling_window"
            n_feature_window_samples = (
                total_training_steps // self.feature_sampling_window
            )
            print(
                f"n_tokens_per_feature_sampling_window (millions): {(self.feature_sampling_window * self.context_size * self.train_batch_size) / 10 ** 6}"
            )
            print(
                f"n_tokens_per_dead_feature_window (millions): {(self.dead_feature_window * self.context_size * self.train_batch_size) / 10 ** 6}"
            )
            print(
                f"We will reset the sparsity calculation {n_feature_window_samples} times."
            )
            # print("Number tokens in dead feature calculation window: ", self.dead_feature_window * self.train_batch_size)
            print(
                f"Number tokens in sparsity calculation window: {self.feature_sampling_window * self.train_batch_size:.2e}"
            )

        if not isinstance(self.use_ghost_grads, list) and self.use_ghost_grads:
            print("Using Ghost Grads.")

    def get_checkpoints_by_step(self) -> tuple[dict[int, str], bool]:
        """
        Returns (dict, is_done)
        where dict is [steps] = path
        for each checkpoint, and
        is_done is True if there is a "final_{steps}" checkpoint
        """
        is_done = False
        checkpoints = [
            f
            for f in os.listdir(self.checkpoint_path)
            if os.path.isdir(os.path.join(self.checkpoint_path, f))
        ]
        mapped_to_steps = {}
        for c in checkpoints:
            try:
                steps = int(c)
            except ValueError:
                if c.startswith("final"):
                    steps = int(c.split("_")[1])
                    is_done = True
                else:
                    continue  # ignore this directory
            full_path = os.path.join(self.checkpoint_path, c)
            mapped_to_steps[steps] = full_path
        return mapped_to_steps, is_done

    def get_resume_checkpoint_path(self) -> str:
        """
        Gets the checkpoint path with the most steps
        raises StopIteration if the model is done (there is a final_{steps} directoryh
        raises FileNotFoundError if there are no checkpoints found
        """
        mapped_to_steps, is_done = self.get_checkpoints_by_step()
        if is_done:
            raise StopIteration("Finished training model")
        if len(mapped_to_steps) == 0:
            raise FileNotFoundError("no checkpoints available to resume from")
        else:
            max_step = max(list(mapped_to_steps.keys()))
            checkpoint_dir = mapped_to_steps[max_step]
            print(f"resuming from step {max_step} at path {checkpoint_dir}")
            return mapped_to_steps[max_step]


@dataclass
class CacheActivationsRunnerConfig:
    """
    Configuration for caching activations of an LLM.
    """

    # Data Generating Function (Model + Training Distibuion)
    model_name: str = "gelu-2l"
    model_class_name: str = "HookedTransformer"
    hook_point: str = "blocks.{layer}.hook_mlp_out"
    hook_point_layer: int | list[int] = 0
    hook_point_head_index: Optional[int] = None
    dataset_path: str = "NeelNanda/c4-tokenized-2b"
    is_dataset_tokenized: bool = True
    context_size: int = 128
    cached_activations_path: Optional[str] = (
        None  # Defaults to "activations/{dataset}/{model}/{full_hook_name}_{hook_point_head_index}"
    )

    # SAE Parameters
    d_in: int = 512

    # Activation Store Parameters
    n_batches_in_buffer: int = 20
    training_tokens: int = 2_000_000
    store_batch_size: int = 32
    train_batch_size: int = 4096

    # Misc
    device: str | torch.device = "cpu"
    seed: int = 42
    dtype: str | torch.dtype = "float32"
    prepend_bos: bool = True

    # Activation caching stuff
    shuffle_every_n_buffers: int = 10
    n_shuffles_with_last_section: int = 10
    n_shuffles_in_entire_dir: int = 10
    n_shuffles_final: int = 100
    model_kwargs: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        # Autofill cached_activations_path unless the user overrode it
        if self.cached_activations_path is None:
            self.cached_activations_path = _default_cached_activations_path(
                self.dataset_path,
                self.model_name,
                self.hook_point,
                self.hook_point_head_index,
            )


def _default_cached_activations_path(
    dataset_path: str,
    model_name: str,
    hook_point: str,
    hook_point_head_index: int | None,
) -> str:
    path = f"activations/{dataset_path.replace('/', '_')}/{model_name.replace('/', '_')}/{hook_point}"
    if hook_point_head_index is not None:
        path += f"_{hook_point_head_index}"
    return path



import pandas as pd
import torch
from tqdm import tqdm
from transformer_lens import HookedTransformer

from sae_lens.training.sparse_autoencoder import SparseAutoencoder


@torch.no_grad()
def get_feature_property_df(
    sparse_autoencoder: SparseAutoencoder, feature_sparsity: torch.Tensor
):
    """
    feature_property_df = get_feature_property_df(sparse_autoencoder, log_feature_density.cpu())
    """

    W_dec_normalized = (
        sparse_autoencoder.W_dec.cpu()
    )  # / sparse_autoencoder.W_dec.cpu().norm(dim=-1, keepdim=True)
    W_enc_normalized = (
        sparse_autoencoder.W_enc.cpu()
        / sparse_autoencoder.W_enc.cpu().norm(dim=-1, keepdim=True)
    ).T

    d_e_projection = (W_dec_normalized * W_enc_normalized).sum(-1)
    b_dec_projection = sparse_autoencoder.b_dec.cpu() @ W_dec_normalized.T

    temp_df = pd.DataFrame(
        {
            "log_feature_sparsity": feature_sparsity + 1e-10,
            "d_e_projection": d_e_projection,
            # "d_e_projection_normalized": d_e_projection_normalized,
            "b_enc": sparse_autoencoder.b_enc.detach().cpu(),
            "b_dec_projection": b_dec_projection,
            "feature": list(range(sparse_autoencoder.cfg.d_sae)),  # type: ignore
            "dead_neuron": (feature_sparsity < -9).cpu(),
        }
    )

    return temp_df


@torch.no_grad()
def get_stats_df(projection: torch.Tensor):
    """
    Returns a dataframe with the mean, std, skewness and kurtosis of the projection
    """
    mean = projection.mean(dim=1, keepdim=True)
    diffs = projection - mean
    var = (diffs**2).mean(dim=1, keepdim=True)
    std = torch.pow(var, 0.5)
    zscores = diffs / std
    skews = torch.mean(torch.pow(zscores, 3.0), dim=1)
    kurtosis = torch.mean(torch.pow(zscores, 4.0), dim=1)

    stats_df = pd.DataFrame(
        {
            "feature": range(len(skews)),
            "mean": mean.numpy().squeeze(),
            "std": std.numpy().squeeze(),
            "skewness": skews.numpy(),
            "kurtosis": kurtosis.numpy(),
        }
    )

    return stats_df


@torch.no_grad()
def get_all_stats_dfs(
    gpt2_small_sparse_autoencoders: dict[str, SparseAutoencoder],  # [hook_point, sae]
    gpt2_small_sae_sparsities: dict[str, torch.Tensor],  # [hook_point, sae]
    model: HookedTransformer,
    cosine_sim: bool = False,
):
    stats_dfs = []
    pbar = tqdm(gpt2_small_sparse_autoencoders.keys())
    for key in pbar:
        layer = int(key.split(".")[1])
        sparse_autoencoder = gpt2_small_sparse_autoencoders[key]
        pbar.set_description(f"Processing layer {sparse_autoencoder.cfg.hook_point}")
        W_U_stats_df_dec, _ = get_W_U_W_dec_stats_df(
            sparse_autoencoder.W_dec.cpu(), model, cosine_sim
        )
        log_feature_sparsity = gpt2_small_sae_sparsities[key].detach().cpu()
        W_U_stats_df_dec["log_feature_sparsity"] = log_feature_sparsity
        W_U_stats_df_dec["layer"] = layer + (1 if "post" in key else 0)
        stats_dfs.append(W_U_stats_df_dec)

    W_U_stats_df_dec_all_layers = pd.concat(stats_dfs, axis=0)
    return W_U_stats_df_dec_all_layers


@torch.no_grad()
def get_W_U_W_dec_stats_df(
    W_dec: torch.Tensor, model: HookedTransformer, cosine_sim: bool = False
) -> tuple[pd.DataFrame, torch.Tensor]:
    W_U = model.W_U.detach().cpu()
    if cosine_sim:
        W_U = W_U / W_U.norm(dim=0, keepdim=True)
    dec_projection_onto_W_U = W_dec @ W_U
    W_U_stats_df = get_stats_df(dec_projection_onto_W_U)
    return W_U_stats_df, dec_projection_onto_W_U



import pickle
from typing import Tuple

import torch


class BackwardsCompatibleUnpickler(pickle.Unpickler):
    """
    An Unpickler that can load files saved before the "sae_lens" package namechange
    """

    def find_class(self, module: str, name: str):
        module = module.replace("sae_training", "sae_lens.training")
        return super().find_class(module, name)


class BackwardsCompatiblePickleClass:
    Unpickler = BackwardsCompatibleUnpickler


def shuffle_activations_pairwise(datapath: str, buffer_idx_range: Tuple[int, int]):
    """
    Shuffles two buffers on disk.
    """
    assert (
        buffer_idx_range[0] < buffer_idx_range[1] - 1
    ), "buffer_idx_range[0] must be smaller than buffer_idx_range[1] by at least 1"

    buffer_idx1 = torch.randint(buffer_idx_range[0], buffer_idx_range[1], (1,)).item()
    buffer_idx2 = torch.randint(buffer_idx_range[0], buffer_idx_range[1], (1,)).item()
    while buffer_idx1 == buffer_idx2:  # Make sure they're not the same
        buffer_idx2 = torch.randint(
            buffer_idx_range[0], buffer_idx_range[1], (1,)
        ).item()

    buffer1 = torch.load(f"{datapath}/{buffer_idx1}.pt")
    buffer2 = torch.load(f"{datapath}/{buffer_idx2}.pt")
    joint_buffer = torch.cat([buffer1, buffer2])

    # Shuffle them
    joint_buffer = joint_buffer[torch.randperm(joint_buffer.shape[0])]
    shuffled_buffer1 = joint_buffer[: buffer1.shape[0]]
    shuffled_buffer2 = joint_buffer[buffer1.shape[0] :]

    # Save them back
    torch.save(shuffled_buffer1, f"{datapath}/{buffer_idx1}.pt")
    torch.save(shuffled_buffer2, f"{datapath}/{buffer_idx2}.pt")



import os
import pickle
import random
import signal
from dataclasses import dataclass, field, fields
from typing import Any, Optional, cast

import numpy as np
import torch
import wandb
from safetensors.torch import save_file
from torch.optim import Adam, Optimizer
from torch.optim.lr_scheduler import LRScheduler
from tqdm import tqdm
from transformer_lens.hook_points import HookedRootModule

from sae_lens import __version__
from sae_lens.training.activations_store import ActivationsStore, HfDataset
from sae_lens.training.config import LanguageModelSAERunnerConfig
from sae_lens.training.evals import run_evals
from sae_lens.training.geometric_median import compute_geometric_median
from sae_lens.training.optim import get_scheduler
from sae_lens.training.sae_group import SparseAutoencoderDictionary
from sae_lens.training.sparse_autoencoder import (
    SAE_CFG_PATH,
    SAE_WEIGHTS_PATH,
    SPARSITY_PATH,
    SparseAutoencoder,
)

# used to map between parameters which are updated during finetuning and the config str.
FINETUNING_PARAMETERS = {
    "scale": ["scaling_factor"],
    "decoder": ["scaling_factor", "W_dec", "b_dec"],
    "unrotated_decoder": ["scaling_factor", "b_dec"],
}


def _log_feature_sparsity(
    feature_sparsity: torch.Tensor, eps: float = 1e-10
) -> torch.Tensor:
    return torch.log10(feature_sparsity + eps).detach().cpu()


@dataclass
class SAETrainContext:
    """
    Context to track during training for a single SAE
    """

    act_freq_scores: torch.Tensor
    n_forward_passes_since_fired: torch.Tensor
    n_frac_active_tokens: int
    optimizer: Optimizer
    scheduler: LRScheduler
    finetuning: bool = False

    @property
    def feature_sparsity(self) -> torch.Tensor:
        return self.act_freq_scores / self.n_frac_active_tokens

    @property
    def log_feature_sparsity(self) -> torch.Tensor:
        return _log_feature_sparsity(self.feature_sparsity)

    def begin_finetuning(self, sae: SparseAutoencoder):

        # finetuning method should be set in the config
        # if not, then we don't finetune
        if not isinstance(sae.cfg.finetuning_method, str):
            return

        for name, param in sae.named_parameters():
            if name in FINETUNING_PARAMETERS[sae.cfg.finetuning_method]:
                param.requires_grad = True
            else:
                param.requires_grad = False

        self.finetuning = True

    def state_dict(self) -> dict[str, torch.Tensor]:
        state_dict = {}
        for attr in fields(self):
            value = getattr(self, attr.name)
            # serializable fields
            if hasattr(value, "state_dict"):
                state_dict[attr.name] = value.state_dict()
            else:
                state_dict[attr.name] = value
        return state_dict

    @classmethod
    def load(cls, path: str, sae: SparseAutoencoder, total_training_steps: int):
        with open(path, "rb") as f:
            state_dict = pickle.load(f)
        attached_ctx = _build_train_context(
            sae=sae, total_training_steps=total_training_steps
        )
        for attr in fields(attached_ctx):
            value = getattr(attached_ctx, attr.name)
            # optimizer and scheduler, this attaches them properly
            if hasattr(value, "state_dict"):
                value.load_state_dict(state_dict[attr.name])
                state_dict[attr.name] = value
        ctx = cls(**state_dict)  # pyright: ignore [reportArgumentType]
        # if fine tuning, we need to set sae requires grad properly
        if ctx.finetuning:
            ctx.begin_finetuning(sae=sae)
        return ctx

    def save(self, path: str):
        with open(path, "wb") as f:
            pickle.dump(self.state_dict(), f)


@dataclass
class SAETrainingRunState:
    """
    Training run state for all SAES
    includes n_training_steps
    n_training_tokens
    started_fine_tuning
    and rng states
    """

    n_training_steps: int = 0
    n_training_tokens: int = 0
    started_fine_tuning: bool = False
    checkpoint_paths: list[str] = field(default_factory=list)
    torch_state: Optional[torch.Tensor] = None
    torch_cuda_state: Optional[list[torch.Tensor]] = None
    numpy_state: Optional[
        dict[str, Any]
        | tuple[str, np.ndarray[Any, np.dtype[np.uint32]], int, int, float]
    ] = None
    random_state: Optional[Any] = None

    def __post_init__(self):
        if self.torch_state is None:
            self.torch_state = torch.get_rng_state()
        if self.torch_cuda_state is None:
            self.torch_cuda_state = torch.cuda.get_rng_state_all()
        if self.numpy_state is None:
            self.numpy_state = np.random.get_state()
        if self.random_state is None:
            self.random_state = random.getstate()

    def set_random_state(self):
        assert self.torch_state is not None
        torch.random.set_rng_state(self.torch_state)
        assert self.torch_cuda_state is not None
        torch.cuda.set_rng_state_all(self.torch_cuda_state)
        assert self.numpy_state is not None
        np.random.set_state(self.numpy_state)
        assert self.random_state is not None
        random.setstate(self.random_state)

    @classmethod
    def load(cls, path: str):
        with open(path, "rb") as f:
            attr_dict = pickle.load(f)
        return cls(**attr_dict)

    def save(self, path: str):
        attr_dict = {**self.__dict__}
        with open(path, "wb") as f:
            pickle.dump(attr_dict, f)


@dataclass
class TrainSAEGroupOutput:
    sae_group: SparseAutoencoderDictionary
    checkpoint_paths: list[str]
    log_feature_sparsities: dict[str, torch.Tensor]


def train_sae_on_language_model(
    model: HookedRootModule,
    sae_group: SparseAutoencoderDictionary,
    activation_store: ActivationsStore,
    batch_size: int = 1024,
    n_checkpoints: int = 0,
    feature_sampling_window: int = 1000,  # how many training steps between resampling the features / considiring neurons dead
    dead_feature_threshold: float = 1e-8,  # how infrequently a feature has to be active to be considered dead
    use_wandb: bool = False,
    wandb_log_frequency: int = 50,
) -> SparseAutoencoderDictionary:
    """
    @deprecated Use `train_sae_group_on_language_model` instead. This method is kept for backward compatibility.
    """
    return train_sae_group_on_language_model(
        model=model,
        sae_group=sae_group,
        activation_store=activation_store,
        batch_size=batch_size,
        n_checkpoints=n_checkpoints,
        feature_sampling_window=feature_sampling_window,
        use_wandb=use_wandb,
        wandb_log_frequency=wandb_log_frequency,
    ).sae_group


def get_total_training_tokens(sae_group: SparseAutoencoderDictionary) -> int:
    return sae_group.cfg.training_tokens + sae_group.cfg.finetuning_tokens


def train_sae_group_on_language_model(
    model: HookedRootModule,
    sae_group: SparseAutoencoderDictionary,
    activation_store: ActivationsStore,
    train_contexts: Optional[dict[str, SAETrainContext]] = None,
    training_run_state: Optional[SAETrainingRunState] = None,
    batch_size: int = 1024,
    n_checkpoints: int = 0,
    feature_sampling_window: int = 1000,  # how many training steps between resampling the features / considiring neurons dead
    use_wandb: bool = False,
    wandb_log_frequency: int = 50,
) -> TrainSAEGroupOutput:
    total_training_tokens = get_total_training_tokens(sae_group=sae_group)
    _update_sae_lens_training_version(sae_group)
    total_training_steps = total_training_tokens // batch_size

    checkpoint_thresholds = []
    if n_checkpoints > 0:
        checkpoint_thresholds = list(
            range(0, total_training_tokens, total_training_tokens // n_checkpoints)
        )[1:]

    all_layers = sae_group.cfg.hook_point_layer
    if not isinstance(all_layers, list):
        all_layers = [all_layers]

    pbar = tqdm(total=total_training_tokens, desc="Training SAE")

    # not resuming
    if training_run_state is None and train_contexts is None:
        train_contexts = {
            name: _build_train_context(sae, total_training_steps)
            for name, sae in sae_group.autoencoders.items()
        }
        training_run_state = SAETrainingRunState()
        _init_sae_group_b_decs(sae_group, activation_store, all_layers)
    # resuming
    else:
        if train_contexts is None:
            raise ValueError(
                "train_contexts is None, when resuming, pass in training_run_state and train_contexts"
            )
        if training_run_state is None:
            raise ValueError(
                "training_run_state is None, when resuming, pass in training_run_state and train_contexts"
            )
        pbar.update(training_run_state.n_training_tokens)
        training_run_state.set_random_state()

    class InterruptedException(Exception):
        pass

    def interrupt_callback(sig_num: Any, stack_frame: Any):
        raise InterruptedException()

    try:
        # signal handlers (if preempted)
        signal.signal(signal.SIGINT, interrupt_callback)
        signal.signal(signal.SIGTERM, interrupt_callback)

        while training_run_state.n_training_tokens < total_training_tokens:
            # Do a training step.
            layer_acts = activation_store.next_batch()
            training_run_state.n_training_tokens += batch_size

            mse_losses: list[torch.Tensor] = []
            l1_losses: list[torch.Tensor] = []

            for name, sparse_autoencoder in sae_group.autoencoders.items():
                ctx = train_contexts[name]
                wandb_suffix = _wandb_log_suffix(sae_group.cfg, sparse_autoencoder.cfg)
                step_output = _train_step(
                    sparse_autoencoder=sparse_autoencoder,
                    layer_acts=layer_acts,
                    ctx=ctx,
                    feature_sampling_window=feature_sampling_window,
                    use_wandb=use_wandb,
                    n_training_steps=training_run_state.n_training_steps,
                    all_layers=all_layers,
                    batch_size=batch_size,
                    wandb_suffix=wandb_suffix,
                )
                mse_losses.append(step_output.mse_loss)
                l1_losses.append(step_output.l1_loss)
                if use_wandb:
                    with torch.no_grad():
                        if (
                            training_run_state.n_training_steps + 1
                        ) % wandb_log_frequency == 0:
                            wandb.log(
                                _build_train_step_log_dict(
                                    sparse_autoencoder,
                                    step_output,
                                    ctx,
                                    wandb_suffix,
                                    training_run_state.n_training_tokens,
                                ),
                                step=training_run_state.n_training_steps,
                            )

                        # record loss frequently, but not all the time.
                        if (training_run_state.n_training_steps + 1) % (
                            wandb_log_frequency * 10
                        ) == 0:
                            sparse_autoencoder.eval()
                            run_evals(
                                sparse_autoencoder,
                                activation_store,
                                model,
                                training_run_state.n_training_steps,
                                suffix=wandb_suffix,
                            )
                            sparse_autoencoder.train()

            # checkpoint if at checkpoint frequency
            if (
                checkpoint_thresholds
                and training_run_state.n_training_tokens > checkpoint_thresholds[0]
            ):
                _save_checkpoint(
                    sae_group,
                    activation_store=activation_store,
                    train_contexts=train_contexts,
                    training_run_state=training_run_state,
                    checkpoint_name=training_run_state.n_training_tokens,
                )
                checkpoint_thresholds.pop(0)

            ###############

            training_run_state.n_training_steps += 1
            pbar.set_description(
                f"{training_run_state.n_training_steps}| MSE Loss {torch.stack(mse_losses).mean().item():.3f} | L1 {torch.stack(l1_losses).mean().item():.3f}"
            )
            pbar.update(batch_size)

            ### If n_training_tokens > sae_group.cfg.training_tokens, then we should switch to fine-tuning (if we haven't already)
            if (not training_run_state.started_fine_tuning) and (
                training_run_state.n_training_tokens > sae_group.cfg.training_tokens
            ):
                training_run_state.started_fine_tuning = True
                for name, sparse_autoencoder in sae_group.autoencoders.items():
                    ctx = train_contexts[name]
                    # this should turn grads on for the scaling factor and other parameters.
                    ctx.begin_finetuning(sae_group.autoencoders[name])
    except (KeyboardInterrupt, InterruptedException):
        print("interrupted, saving progress")
        checkpoint_name = training_run_state.n_training_tokens
        _save_checkpoint(
            sae_group,
            activation_store=activation_store,
            train_contexts=train_contexts,
            training_run_state=training_run_state,
            checkpoint_name=checkpoint_name,
        )
        print("done saving")
        raise
    # save final sae group to checkpoints folder
    _save_checkpoint(
        sae_group,
        activation_store=activation_store,
        train_contexts=train_contexts,
        training_run_state=training_run_state,
        checkpoint_name=f"final_{training_run_state.n_training_tokens}",
        wandb_aliases=["final_model"],
    )

    log_feature_sparsities = {
        name: ctx.log_feature_sparsity for name, ctx in train_contexts.items()
    }

    return TrainSAEGroupOutput(
        sae_group=sae_group,
        checkpoint_paths=training_run_state.checkpoint_paths,
        log_feature_sparsities=log_feature_sparsities,
    )


def _wandb_log_suffix(cfg: Any, hyperparams: Any):
    # Create a mapping from cfg list keys to their corresponding hyperparams attributes
    key_mapping = {
        "hook_point_layer": "layer",
        "l1_coefficient": "coeff",
        "lp_norm": "l",
        "lr": "lr",
    }

    # Generate the suffix by iterating over the keys that have list values in cfg
    suffix = "".join(
        f"_{key_mapping.get(key, key)}{getattr(hyperparams, key, '')}"
        for key, value in vars(cfg).items()
        if isinstance(value, list)
    )
    return suffix


def _build_train_context(
    sae: SparseAutoencoder, total_training_steps: int
) -> SAETrainContext:
    assert not isinstance(sae.cfg.lr, list), "lr must not be a list for a single SAE"
    assert not isinstance(
        sae.cfg.lr_end, list
    ), "lr_end must not be a list for a single SAE"
    assert not isinstance(
        sae.cfg.lr_scheduler_name, list
    ), "lr_scheduler_name must not be a list for a single SAE"
    assert not isinstance(
        sae.cfg.lr_warm_up_steps, list
    ), "lr_warm_up_steps must not be a list for a single SAE"
    assert not isinstance(
        sae.cfg.lr_decay_steps, list
    ), "lr_decay_steps must not be a list for a single SAE"
    assert not isinstance(
        sae.cfg.n_restart_cycles, list
    ), "n_restart_cycles must not be a list for a single SAE"

    act_freq_scores = torch.zeros(
        cast(int, sae.cfg.d_sae),
        device=sae.cfg.device,
    )
    n_forward_passes_since_fired = torch.zeros(
        cast(int, sae.cfg.d_sae),
        device=sae.cfg.device,
    )
    n_frac_active_tokens = 0

    # we don't train the scaling factor (initially)
    # set requires grad to false for the scaling factor
    for name, param in sae.named_parameters():
        if "scaling_factor" in name:
            param.requires_grad = False

    optimizer = Adam(
        sae.parameters(),
        lr=sae.cfg.lr,
        betas=(
            sae.cfg.adam_beta1,  # type: ignore
            sae.cfg.adam_beta2,  # type: ignore
        ),
    )
    assert sae.cfg.lr_end is not None  # this is set in config post-init
    scheduler = get_scheduler(
        sae.cfg.lr_scheduler_name,
        lr=sae.cfg.lr,
        optimizer=optimizer,
        warm_up_steps=sae.cfg.lr_warm_up_steps,
        decay_steps=sae.cfg.lr_decay_steps,
        training_steps=total_training_steps,
        lr_end=sae.cfg.lr_end,
        num_cycles=sae.cfg.n_restart_cycles,
    )

    return SAETrainContext(
        act_freq_scores=act_freq_scores,
        n_forward_passes_since_fired=n_forward_passes_since_fired,
        n_frac_active_tokens=n_frac_active_tokens,
        optimizer=optimizer,
        scheduler=scheduler,
    )


def _init_sae_group_b_decs(
    sae_group: SparseAutoencoderDictionary,
    activation_store: ActivationsStore,
    all_layers: list[int],
) -> None:
    """
    extract all activations at a certain layer and use for sae b_dec initialization
    """
    geometric_medians = {}
    for _, sae in sae_group:
        hyperparams = sae.cfg
        sae_layer_id = all_layers.index(sae.hook_point_layer)
        if hyperparams.b_dec_init_method == "geometric_median":
            layer_acts = activation_store.storage_buffer.detach()[:, sae_layer_id, :]
            # get geometric median of the activations if we're using those.
            if sae_layer_id not in geometric_medians:
                median = compute_geometric_median(
                    layer_acts,
                    maxiter=100,
                ).median
                geometric_medians[sae_layer_id] = median
            sae.initialize_b_dec_with_precalculated(geometric_medians[sae_layer_id])
        elif hyperparams.b_dec_init_method == "mean":
            layer_acts = activation_store.storage_buffer.detach().cpu()[
                :, sae_layer_id, :
            ]
            sae.initialize_b_dec_with_mean(layer_acts)


def _update_sae_lens_training_version(sae_group: SparseAutoencoderDictionary) -> None:
    """
    Make sure we record the version of SAELens used for the training run
    """
    sae_group.cfg.sae_lens_training_version = __version__
    for sae in sae_group.autoencoders.values():
        sae.cfg.sae_lens_training_version = __version__


@dataclass
class TrainStepOutput:
    sae_in: torch.Tensor
    sae_out: torch.Tensor
    feature_acts: torch.Tensor
    loss: torch.Tensor
    mse_loss: torch.Tensor
    l1_loss: torch.Tensor
    ghost_grad_loss: torch.Tensor
    ghost_grad_neuron_mask: torch.Tensor


def _train_step(
    sparse_autoencoder: SparseAutoencoder,
    layer_acts: torch.Tensor,
    ctx: SAETrainContext,
    feature_sampling_window: int,  # how many training steps between resampling the features / considiring neurons dead
    use_wandb: bool,
    n_training_steps: int,
    all_layers: list[int],
    batch_size: int,
    wandb_suffix: str,
) -> TrainStepOutput:
    assert sparse_autoencoder.cfg.d_sae is not None  # keep pyright happy
    layer_id = all_layers.index(sparse_autoencoder.hook_point_layer)
    sae_in = layer_acts[:, layer_id, :]

    sparse_autoencoder.train()
    # Make sure the W_dec is still zero-norm
    if sparse_autoencoder.normalize_sae_decoder:
        sparse_autoencoder.set_decoder_norm_to_unit_norm()

    # log and then reset the feature sparsity every feature_sampling_window steps
    if (n_training_steps + 1) % feature_sampling_window == 0:
        feature_sparsity = ctx.feature_sparsity
        log_feature_sparsity = _log_feature_sparsity(feature_sparsity)

        if use_wandb:
            wandb_histogram = wandb.Histogram(log_feature_sparsity.numpy())
            wandb.log(
                {
                    f"metrics/mean_log10_feature_sparsity{wandb_suffix}": log_feature_sparsity.mean().item(),
                    f"plots/feature_density_line_chart{wandb_suffix}": wandb_histogram,
                    f"sparsity/below_1e-5{wandb_suffix}": (feature_sparsity < 1e-5)
                    .sum()
                    .item(),
                    f"sparsity/below_1e-6{wandb_suffix}": (feature_sparsity < 1e-6)
                    .sum()
                    .item(),
                },
                step=n_training_steps,
            )

        ctx.act_freq_scores = torch.zeros(
            sparse_autoencoder.cfg.d_sae, device=sparse_autoencoder.cfg.device
        )
        ctx.n_frac_active_tokens = 0

    ghost_grad_neuron_mask = (
        ctx.n_forward_passes_since_fired > sparse_autoencoder.cfg.dead_feature_window
    ).bool()

    # Forward and Backward Passes
    (
        sae_out,
        feature_acts,
        loss,
        mse_loss,
        l1_loss,
        ghost_grad_loss,
    ) = sparse_autoencoder(
        sae_in,
        ghost_grad_neuron_mask,
    )
    did_fire = (feature_acts > 0).float().sum(-2) > 0
    ctx.n_forward_passes_since_fired += 1
    ctx.n_forward_passes_since_fired[did_fire] = 0

    with torch.no_grad():
        # Calculate the sparsities, and add it to a list, calculate sparsity metrics
        ctx.act_freq_scores += (feature_acts.abs() > 0).float().sum(0)
        ctx.n_frac_active_tokens += batch_size

    ctx.optimizer.zero_grad()
    loss.backward()
    if sparse_autoencoder.normalize_sae_decoder:
        sparse_autoencoder.remove_gradient_parallel_to_decoder_directions()
    ctx.optimizer.step()
    ctx.scheduler.step()

    return TrainStepOutput(
        sae_in=sae_in,
        sae_out=sae_out,
        feature_acts=feature_acts,
        loss=loss,
        mse_loss=mse_loss,
        l1_loss=l1_loss,
        ghost_grad_loss=ghost_grad_loss,
        ghost_grad_neuron_mask=ghost_grad_neuron_mask,
    )


def _build_train_step_log_dict(
    sparse_autoencoder: SparseAutoencoder,
    output: TrainStepOutput,
    ctx: SAETrainContext,
    wandb_suffix: str,
    n_training_tokens: int,
) -> dict[str, Any]:
    sae_in = output.sae_in
    sae_out = output.sae_out
    feature_acts = output.feature_acts
    mse_loss = output.mse_loss
    l1_loss = output.l1_loss
    ghost_grad_loss = output.ghost_grad_loss
    loss = output.loss
    ghost_grad_neuron_mask = output.ghost_grad_neuron_mask

    # metrics for currents acts
    l0 = (feature_acts > 0).float().sum(-1).mean()
    current_learning_rate = ctx.optimizer.param_groups[0]["lr"]

    per_token_l2_loss = (sae_out - sae_in).pow(2).sum(dim=-1).squeeze()
    total_variance = (sae_in - sae_in.mean(0)).pow(2).sum(-1)
    explained_variance = 1 - per_token_l2_loss / total_variance

    return {
        # losses
        f"losses/mse_loss{wandb_suffix}": mse_loss.item(),
        f"losses/l1_loss{wandb_suffix}": l1_loss.item()
        / sparse_autoencoder.l1_coefficient,  # normalize by l1 coefficient
        f"losses/ghost_grad_loss{wandb_suffix}": ghost_grad_loss.item(),
        f"losses/overall_loss{wandb_suffix}": loss.item(),
        # variance explained
        f"metrics/explained_variance{wandb_suffix}": explained_variance.mean().item(),
        f"metrics/explained_variance_std{wandb_suffix}": explained_variance.std().item(),
        f"metrics/l0{wandb_suffix}": l0.item(),
        # sparsity
        f"sparsity/mean_passes_since_fired{wandb_suffix}": ctx.n_forward_passes_since_fired.mean().item(),
        f"sparsity/dead_features{wandb_suffix}": ghost_grad_neuron_mask.sum().item(),
        f"details/current_learning_rate{wandb_suffix}": current_learning_rate,
        "details/n_training_tokens": n_training_tokens,
    }


ACTIVATION_STORE_PATH = "activation_store.safetensors"
TRAINING_RUN_STATE_PATH = "training_run_state.pkl"
SAE_CONTEXT_PATH = "ctx.safetensors"


def load_checkpoint(
    checkpoint_path: str,
    cfg: LanguageModelSAERunnerConfig,
    model: HookedRootModule,
    batch_size: int,
    dataset: HfDataset | None = None,
) -> tuple[
    SAETrainingRunState,
    ActivationsStore,
    SparseAutoencoderDictionary,
    dict[str, SAETrainContext],
]:
    training_run_state_path = f"{checkpoint_path}/{TRAINING_RUN_STATE_PATH}"
    training_run_state = SAETrainingRunState.load(training_run_state_path)

    activation_store_path = f"{checkpoint_path}/{ACTIVATION_STORE_PATH}"
    activation_store = ActivationsStore.load(
        activation_store_path, model=model, cfg=cfg, dataset=dataset
    )

    sae_group = SparseAutoencoderDictionary.load_from_pretrained(
        checkpoint_path, device=str(cfg.device)
    )

    total_training_steps = get_total_training_tokens(sae_group=sae_group) // batch_size

    train_contexts = {}
    for name, sae in sae_group.autoencoders.items():
        path = f"{checkpoint_path}/{name}"
        ctx_path = f"{path}/{SAE_CONTEXT_PATH}"
        train_contexts[name] = SAETrainContext.load(
            ctx_path, sae=sae, total_training_steps=total_training_steps
        )

    # overwrite sae gruop cfg with our new cfg in case we want to change things
    sae_group.cfg = cfg
    # TODO: individual saes don't get new cfgs, maybe they should idk its messy bc of _init_autoencoders stuff
    return training_run_state, activation_store, sae_group, train_contexts


def _save_checkpoint(
    sae_group: SparseAutoencoderDictionary,
    activation_store: ActivationsStore,
    train_contexts: dict[str, SAETrainContext],
    training_run_state: SAETrainingRunState,
    checkpoint_name: int | str,
    wandb_aliases: list[str] | None = None,
) -> str:

    checkpoint_path = f"{sae_group.cfg.checkpoint_path}/{checkpoint_name}"
    training_run_state.checkpoint_paths.append(checkpoint_path)
    os.makedirs(checkpoint_path, exist_ok=True)

    training_run_state_path = f"{checkpoint_path}/{TRAINING_RUN_STATE_PATH}"
    training_run_state.save(training_run_state_path)
    if sae_group.cfg.log_to_wandb:
        training_run_state_artifact = wandb.Artifact(
            f"{sae_group.get_name()}_training_run_state",
            type="training_run_state",
            metadata=dict(sae_group.cfg.__dict__),
        )
        training_run_state_artifact.add_file(training_run_state_path)
        # TODO: should these have aliases=wandb_aliases?
        wandb.log_artifact(training_run_state_artifact)

    activation_store_path = f"{checkpoint_path}/{ACTIVATION_STORE_PATH}"
    activation_store.save(activation_store_path)
    if sae_group.cfg.log_to_wandb and sae_group.cfg.log_activations_store_to_wandb:
        activation_store_artifact = wandb.Artifact(
            f"{sae_group.get_name()}_activations_store",
            type="activation_store",
            metadata=dict(sae_group.cfg.__dict__),
        )
        activation_store_artifact.add_file(activation_store_path)
        wandb.log_artifact(activation_store_artifact)

    for name, sae in sae_group.autoencoders.items():

        ctx = train_contexts[name]
        path = f"{checkpoint_path}/{name}"
        os.makedirs(path, exist_ok=True)
        ctx_path = f"{path}/{SAE_CONTEXT_PATH}"
        ctx.save(ctx_path)

        if sae.normalize_sae_decoder:
            sae.set_decoder_norm_to_unit_norm()
        sae.save_model(path)
        log_feature_sparsities = {"sparsity": ctx.log_feature_sparsity}

        log_feature_sparsity_path = f"{path}/{SPARSITY_PATH}"
        save_file(log_feature_sparsities, log_feature_sparsity_path)

        if sae_group.cfg.log_to_wandb and os.path.exists(log_feature_sparsity_path):
            model_artifact = wandb.Artifact(
                f"{sae_group.get_name()}",
                type="model",
                metadata=dict(sae_group.cfg.__dict__),
            )
            model_artifact.add_file(f"{path}/{SAE_WEIGHTS_PATH}")
            model_artifact.add_file(f"{path}/{SAE_CFG_PATH}")
            if sae_group.cfg.log_optimizer_state_to_wandb:
                model_artifact.add_file(ctx_path)
            wandb.log_artifact(model_artifact, aliases=wandb_aliases)

            sparsity_artifact = wandb.Artifact(
                f"{sae_group.get_name()}_log_feature_sparsity",
                type="log_feature_sparsity",
                metadata=dict(sae_group.cfg.__dict__),
            )
            sparsity_artifact.add_file(log_feature_sparsity_path)
            wandb.log_artifact(sparsity_artifact)

    return checkpoint_path



from __future__ import annotations

import dataclasses
import gzip
import os
import pickle
from itertools import product
from types import SimpleNamespace
from typing import Iterator

import torch

from sae_lens.training.config import LanguageModelSAERunnerConfig
from sae_lens.training.sparse_autoencoder import SparseAutoencoder
from sae_lens.training.utils import BackwardsCompatibleUnpickler


class SparseAutoencoderDictionary:

    autoencoders: dict[str, SparseAutoencoder]

    def __init__(self, cfg: LanguageModelSAERunnerConfig):
        self.cfg = cfg
        self.autoencoders = {}  # This will store tuples of (instance, hyperparameters)
        self._init_autoencoders(cfg)

    def _init_autoencoders(self, cfg: LanguageModelSAERunnerConfig):
        # Dynamically get all combinations of hyperparameters from cfg
        # Extract all hyperparameter lists from cfg
        hyperparameters = {k: v for k, v in vars(cfg).items() if isinstance(v, list)}
        if len(hyperparameters) > 0:
            keys, values = zip(*hyperparameters.items())
        else:
            keys, values = (), ([()],)  # Ensure product(*values) yields one combination

        # Create all combinations of hyperparameters
        for combination in product(*values):

            params = dict(zip(keys, combination))
            cfg_copy = dataclasses.replace(cfg, **params)

            cfg_copy.__post_init__()
            # Insert the layer into the hookpoint
            cfg_copy.hook_point = cfg_copy.hook_point.format(
                layer=cfg_copy.hook_point_layer
            )

            sae = SparseAutoencoder(cfg_copy)

            sae_name = (
                f"{sae.cfg.model_name.replace('/', '_')}_{sae.cfg.hook_point}_{sae.cfg.d_sae}_"
                + "_".join([f"{k}_{v}" for k, v in zip(keys, combination)])
            )

            # Create and store both the SparseAutoencoder instance and its parameters
            self.autoencoders[sae_name] = sae

    def __getitem__(self, key: str) -> SparseAutoencoder:
        return self.autoencoders[key]

    def __iter__(self) -> Iterator[tuple[str, SparseAutoencoder]]:
        # Make SAEGroup iterable over its SparseAutoencoder instances and their parameters
        for name, sae in self.autoencoders.items():
            yield name, sae  # Yielding as a tuple

    def __len__(self):
        # Return the number of SparseAutoencoder instances
        return len(self.autoencoders)

    def to(self, device: torch.device | str):
        for ae in self.autoencoders.values():
            ae.to(device)

    @classmethod
    def load_from_pretrained_legacy(cls, path: str) -> "SparseAutoencoderDictionary":
        """
        Load function for the model. Loads the model's state_dict and the config used to train it.
        This method can be called directly on the class, without needing an instance.
        """

        # Ensure the file exists
        if not os.path.isfile(path):
            raise FileNotFoundError(f"No file found at specified path: {path}")

        # Load the state dictionary
        if path.endswith(".pt"):
            try:
                # this is hacky, but can't figure out how else to get torch to use our custom unpickler
                fake_pickle = SimpleNamespace()
                fake_pickle.Unpickler = BackwardsCompatibleUnpickler
                fake_pickle.__name__ = pickle.__name__

                if torch.cuda.is_available():
                    group = torch.load(
                        path,
                        pickle_module=fake_pickle,
                    )
                else:
                    map_loc = "mps" if torch.backends.mps.is_available() else "cpu"
                    group = torch.load(
                        path, pickle_module=fake_pickle, map_location=map_loc
                    )
                    if isinstance(group, dict):
                        group["cfg"].device = map_loc
                    else:
                        group.cfg.device = map_loc
            except Exception as e:
                raise IOError(f"Error loading the state dictionary from .pt file: {e}")

        elif path.endswith(".pkl.gz"):
            try:
                with gzip.open(path, "rb") as f:
                    group = pickle.load(f)
            except Exception as e:
                raise IOError(
                    f"Error loading the state dictionary from .pkl.gz file: {e}"
                )
        elif path.endswith(".pkl"):
            try:
                with open(path, "rb") as f:
                    group = pickle.load(f)
            except Exception as e:
                raise IOError(f"Error loading the state dictionary from .pkl file: {e}")
        else:
            raise ValueError(
                f"Unexpected file extension: {path}, supported extensions are .pt, .pkl, and .pkl.gz"
            )

        # handle loading old autoencoders where before SAEGroup existed, where we just save a dict
        if isinstance(group, dict):
            cfg = group["cfg"]
            # need to add fields to old configs
            if not hasattr(cfg, "model_kwargs"):
                cfg.model_kwargs = {}
            if not hasattr(cfg, "sae_lens_version"):
                cfg.sae_lens_version = "0.0.0"
            if not hasattr(cfg, "sae_lens_training_version"):
                cfg.sae_lens_training_version = "0.0.0"
            sparse_autoencoder = SparseAutoencoder(cfg=cfg)
            # add dummy scaling factor to the state dict
            group["state_dict"]["scaling_factor"] = torch.ones(
                cfg.d_sae, dtype=cfg.dtype, device=cfg.device
            )
            sparse_autoencoder.load_state_dict(group["state_dict"])
            group = cls(cfg)
            for key in group.autoencoders:
                group.autoencoders[key] = sparse_autoencoder

        if not isinstance(group, cls):
            raise ValueError("The loaded object is not a valid SAEGroup")

        return group

    @classmethod
    def load_from_pretrained(
        cls, path: str, device: str = "cpu"
    ) -> "SparseAutoencoderDictionary":

        autoencoders = {}

        # check if there any folders inside the current path
        folders_in_current_path = [
            f for f in os.listdir(path) if os.path.isdir(os.path.join(path, f))
        ]
        if len(folders_in_current_path) == 0:
            # sae_name is the folder name
            sae_name = os.path.basename(path)
            autoencoders[sae_name] = SparseAutoencoder.load_from_pretrained(
                path,
                device,
            )
        else:  # we have a list of SAE folders
            for sae_name in os.listdir(path):
                if os.path.isdir(os.path.join(path, sae_name)):
                    autoencoders[sae_name] = SparseAutoencoder.load_from_pretrained(
                        os.path.join(path, sae_name),
                        device,
                    )

        # Create the SAEGroup object

        # loaded SAE groups will not contain a cfg that matches the original.
        # TODO: figure out how to handle this
        sae_name = next(iter(autoencoders.keys()))
        sae_group = cls(autoencoders[sae_name].cfg)
        sae_group.autoencoders = autoencoders
        return sae_group

    def save_saes(self, path: str):
        """
        Basic save function for the model. Saves the model's state_dict and the config used to train it.
        """

        # check if path exists
        folder = os.path.dirname(path)
        os.makedirs(folder, exist_ok=True)

        if len(self.autoencoders) == 1:
            sae: SparseAutoencoder = next(iter(self))[1]
            sae.save_model(path)
        else:
            for i, autoencoder in self.autoencoders.items():
                subfolder_name = f"{path}/{i}"
                os.makedirs(subfolder_name, exist_ok=True)
                autoencoder.save_model(f"{path}/{i}")

    def get_name(self):
        sae_name = f"sae_group_{self.cfg.model_name.replace('/', '_')}_{self.cfg.hook_point}_{self.cfg.d_sae}"
        return sae_name

    def eval(self):
        for ae in self.autoencoders.values():
            ae.eval()

    def train(self):
        for ae in self.autoencoders.values():
            ae.train()



__version__ = "1.4.0"

from .training.activations_store import ActivationsStore
from .training.cache_activations_runner import cache_activations_runner
from .training.config import CacheActivationsRunnerConfig, LanguageModelSAERunnerConfig
from .training.evals import run_evals
from .training.lm_runner import language_model_sae_runner
from .training.sae_group import SparseAutoencoderDictionary
from .training.session_loader import LMSparseAutoencoderSessionloader
from .training.sparse_autoencoder import SparseAutoencoder
from .training.train_sae_on_language_model import train_sae_group_on_language_model

__all__ = [
    "LanguageModelSAERunnerConfig",
    "CacheActivationsRunnerConfig",
    "LMSparseAutoencoderSessionloader",
    "SparseAutoencoder",
    "SparseAutoencoderDictionary",
    "run_evals",
    "language_model_sae_runner",
    "cache_activations_runner",
    "ActivationsStore",
    "train_sae_group_on_language_model",
]



from typing import Callable

import torch


def get_activation_fn(activation_fn: str) -> Callable[[torch.Tensor], torch.Tensor]:
    if activation_fn == "relu":
        return torch.nn.ReLU()
    elif activation_fn == "tanh-relu":
        return tanh_relu
    else:
        raise ValueError(f"Unknown activation function: {activation_fn}")


def tanh_relu(input: torch.Tensor) -> torch.Tensor:
    input = torch.relu(input)
    input = torch.tanh(input)
    return input



from dataclasses import dataclass
from typing import Any, cast

import einops
import torch
import wandb

from sae_lens.training.sparse_autoencoder import SparseAutoencoder
from sae_lens.training.toy_models import Config as ToyConfig
from sae_lens.training.toy_models import Model as ToyModel
from sae_lens.training.train_sae_on_toy_model import train_toy_sae


@dataclass
class SAEToyModelRunnerConfig:
    # ReLu Model Parameters
    n_features: int = 5
    n_hidden: int = 2
    n_correlated_pairs: int = 0
    n_anticorrelated_pairs: int = 0
    feature_probability: float = 0.025
    model_training_steps: int = 10_000

    # SAE Parameters
    d_sae: int = 5

    # Training Parameters
    l1_coefficient: float = 1e-3
    lr: float = 3e-4
    train_batch_size: int = 1024
    b_dec_init_method: str = "geometric_median"

    # Sparsity / Dead Feature Handling
    use_ghost_grads: bool = (
        False  # not currently implemented, but SAE class expects it.
    )
    feature_sampling_window: int = 100
    dead_feature_window: int = 100  # unless this window is larger feature sampling,
    dead_feature_threshold: float = 1e-8

    # Activation Store Parameters
    total_training_tokens: int = 25_000

    # WANDB
    log_to_wandb: bool = True
    wandb_project: str = "mats_sae_training_toy_model"
    wandb_entity: str | None = None
    wandb_log_frequency: int = 50

    # Misc
    device: str = "cpu"
    seed: int = 42
    checkpoint_path: str = "checkpoints"
    dtype: torch.dtype = torch.float32

    def __post_init__(self):
        self.d_in = self.n_hidden  # hidden for the ReLu model is the input for the SAE


def toy_model_sae_runner(cfg: SAEToyModelRunnerConfig):
    """
    A runner for training an SAE on a toy model.
    """
    # Toy Model Config
    toy_model_cfg = ToyConfig(
        n_instances=1,  # Not set up to train > 1 SAE so shouldn't do > 1 model.
        n_features=cfg.n_features,
        n_hidden=cfg.n_hidden,
        n_correlated_pairs=cfg.n_correlated_pairs,
        n_anticorrelated_pairs=cfg.n_anticorrelated_pairs,
    )

    # Initialize Toy Model
    model = ToyModel(
        cfg=toy_model_cfg,
        device=cfg.device,
        feature_probability=cfg.feature_probability,
    )

    # Train the Toy Model
    model.optimize(steps=cfg.model_training_steps)

    # Generate Training Data
    batch = model.generate_batch(cfg.total_training_tokens)
    hidden = einops.einsum(
        batch,
        model.W,
        "batch_size instances features, instances hidden features -> batch_size instances hidden",
    )

    sparse_autoencoder = SparseAutoencoder(
        cast(Any, cfg)  # TODO: the types are broken here
    )  # config has the hyperparameters for the SAE

    if cfg.log_to_wandb:
        wandb.init(project=cfg.wandb_project, config=cast(Any, cfg))

    sparse_autoencoder = train_toy_sae(
        sparse_autoencoder,
        activation_store=hidden.detach().squeeze(),
        batch_size=cfg.train_batch_size,
        feature_sampling_window=cfg.feature_sampling_window,
        dead_feature_threshold=cfg.dead_feature_threshold,
        use_wandb=cfg.log_to_wandb,
        wandb_log_frequency=cfg.wandb_log_frequency,
    )

    if cfg.log_to_wandb:
        wandb.finish()

    return sparse_autoencoder



import math
import os

import torch
from tqdm import tqdm

from sae_lens.training.activations_store import ActivationsStore
from sae_lens.training.config import CacheActivationsRunnerConfig
from sae_lens.training.load_model import load_model
from sae_lens.training.utils import shuffle_activations_pairwise


def cache_activations_runner(cfg: CacheActivationsRunnerConfig):
    model = load_model(
        model_class_name=cfg.model_class_name,
        model_name=cfg.model_name,
        device=cfg.device,
    )
    activations_store = ActivationsStore.from_config(
        model,
        cfg,
    )

    # if the activations directory exists and has files in it, raise an exception
    assert activations_store.cached_activations_path is not None
    if os.path.exists(activations_store.cached_activations_path):
        if len(os.listdir(activations_store.cached_activations_path)) > 0:
            raise Exception(
                f"Activations directory ({activations_store.cached_activations_path}) is not empty. Please delete it or specify a different path. Exiting the script to prevent accidental deletion of files."
            )
    else:
        os.makedirs(activations_store.cached_activations_path)

    print(f"Started caching {cfg.training_tokens} activations")
    tokens_per_buffer = (
        cfg.store_batch_size * cfg.context_size * cfg.n_batches_in_buffer
    )
    n_buffers = math.ceil(cfg.training_tokens / tokens_per_buffer)
    # for i in tqdm(range(n_buffers), desc="Caching activations"):
    for i in range(n_buffers):
        buffer = activations_store.get_buffer(cfg.n_batches_in_buffer)
        torch.save(buffer, f"{activations_store.cached_activations_path}/{i}.pt")
        del buffer

        if i % cfg.shuffle_every_n_buffers == 0 and i > 0:
            # Shuffle the buffers on disk

            # Do random pairwise shuffling between the last shuffle_every_n_buffers buffers
            for _ in range(cfg.n_shuffles_with_last_section):
                shuffle_activations_pairwise(
                    activations_store.cached_activations_path,
                    buffer_idx_range=(i - cfg.shuffle_every_n_buffers, i),
                )

            # Do more random pairwise shuffling between all the buffers
            for _ in range(cfg.n_shuffles_in_entire_dir):
                shuffle_activations_pairwise(
                    activations_store.cached_activations_path,
                    buffer_idx_range=(0, i),
                )

    # More final shuffling (mostly in case we didn't end on an i divisible by shuffle_every_n_buffers)
    if n_buffers > 1:
        for _ in tqdm(range(cfg.n_shuffles_final), desc="Final shuffling"):
            shuffle_activations_pairwise(
                activations_store.cached_activations_path,
                buffer_idx_range=(0, n_buffers),
            )



import os
from typing import Any, Dict, List, Optional, Union, cast

# set TOKENIZERS_PARALLELISM to false to avoid warnings
os.environ["TOKENIZERS_PARALLELISM"] = "false"
import json

import numpy as np
import torch
from matplotlib import colors
from sae_vis.data_config_classes import (
    ActsHistogramConfig,
    Column,
    FeatureTablesConfig,
    LogitsHistogramConfig,
    LogitsTableConfig,
    SaeVisConfig,
    SaeVisLayoutConfig,
    SequencesConfig,
)
from sae_vis.data_storing_fns import SaeVisData
from tqdm import tqdm
from transformer_lens import HookedTransformer

from sae_lens.toolkit.pretrained_saes import load_sparsity
from sae_lens.training.session_loader import LMSparseAutoencoderSessionloader
from sae_lens.training.sparse_autoencoder import SparseAutoencoder

OUT_OF_RANGE_TOKEN = "<|outofrange|>"

BG_COLOR_MAP = colors.LinearSegmentedColormap.from_list(
    "bg_color_map", ["white", "darkorange"]
)

DEFAULT_SPARSITY_THRESHOLD = -5

HTML_ANOMALIES = {
    "âĢĶ": "—",
    "âĢĵ": "–",
    "âĢľ": "“",
    "âĢĿ": "”",
    "âĢĺ": "‘",
    "âĢĻ": "’",
    "âĢĭ": " ",  # todo: this is actually zero width space
    "Ġ": " ",
    "Ċ": "\n",
    "ĉ": "\t",
}


class NpEncoder(json.JSONEncoder):
    def default(self, o: Any):
        if isinstance(o, np.integer):
            return int(o)
        if isinstance(o, np.floating):
            return float(o)
        if isinstance(o, np.ndarray):
            return o.tolist()
        return super(NpEncoder, self).default(o)


class NeuronpediaRunner:

    def __init__(
        self,
        sae_id: str,
        sae_path: str,
        outputs_dir: str,
        sparsity_threshold: int = DEFAULT_SPARSITY_THRESHOLD,
        # token pars
        n_batches_to_sample_from: int = 2**12,
        n_prompts_to_select: int = 4096 * 6,
        # batching
        n_features_at_a_time: int = 128,
        start_batch_inclusive: int = 0,
        end_batch_inclusive: Optional[int] = None,
        # quantiles
        n_quantiles: int = 5,
        top_acts_group_size: int = 20,
        quantile_group_size: int = 5,
    ):

        self.device = "cpu"
        if torch.backends.mps.is_available():
            self.device = "mps"
        elif torch.cuda.is_available():
            self.device = "cuda"

        self.sae_path = sae_path
        self.sparse_autoencoder = SparseAutoencoder.load_from_pretrained(
            self.sae_path, device=self.device
        )
        loader = LMSparseAutoencoderSessionloader(self.sparse_autoencoder.cfg)
        self.model, _, self.activation_store = loader.load_sae_training_group_session()
        self.model_id = self.sparse_autoencoder.cfg.model_name
        self.layer = self.sparse_autoencoder.cfg.hook_point_layer
        self.sae_id = sae_id
        self.sparsity_threshold = sparsity_threshold
        self.n_features_at_a_time = n_features_at_a_time
        self.n_batches_to_sample_from = n_batches_to_sample_from
        self.n_prompts_to_select = n_prompts_to_select
        self.start_batch = start_batch_inclusive
        self.end_batch = end_batch_inclusive
        self.n_quantiles = n_quantiles
        self.top_acts_group_size = top_acts_group_size
        self.quantile_group_size = quantile_group_size

        if not os.path.exists(outputs_dir):
            os.makedirs(outputs_dir)
        self.outputs_dir = outputs_dir

    def get_tokens(
        self,
        n_batches_to_sample_from: int = 2**12,
        n_prompts_to_select: int = 4096 * 6,
    ):
        all_tokens_list = []
        pbar = tqdm(range(n_batches_to_sample_from))
        for _ in pbar:
            batch_tokens = self.activation_store.get_batch_tokens()
            batch_tokens = batch_tokens[torch.randperm(batch_tokens.shape[0])][
                : batch_tokens.shape[0]
            ]
            all_tokens_list.append(batch_tokens)

        all_tokens = torch.cat(all_tokens_list, dim=0)
        all_tokens = all_tokens[torch.randperm(all_tokens.shape[0])]
        return all_tokens[:n_prompts_to_select]

    def round_list(self, to_round: list[float]):
        return list(np.round(to_round, 3))

    def to_str_tokens_safe(
        self,
        vocab_dict: Dict[int, str],
        tokens: Union[int, List[int], torch.Tensor],
    ):
        """
        does to_str_tokens, except handles out of range
        """
        assert self.model is not None
        vocab_max_index = self.model.cfg.d_vocab - 1
        # Deal with the int case separately
        if isinstance(tokens, int):
            if tokens > vocab_max_index:
                return OUT_OF_RANGE_TOKEN
            return vocab_dict[tokens]

        # If the tokens are a (possibly nested) list, turn them into a tensor
        if isinstance(tokens, list):
            tokens = torch.tensor(tokens)

        # Get flattened list of tokens
        str_tokens = [
            (vocab_dict[t] if t <= vocab_max_index else OUT_OF_RANGE_TOKEN)
            for t in tokens.flatten().tolist()
        ]

        # Reshape
        return np.reshape(str_tokens, tokens.shape).tolist()

    def run(self):
        self.n_features = self.sparse_autoencoder.cfg.d_sae
        assert self.n_features is not None

        # if we have feature sparsity, then use it to only generate outputs for non-dead features
        self.target_feature_indexes: list[int] = []
        sparsity = load_sparsity(self.sae_path)
        # convert sparsity to logged sparsity if it's not
        # TODO: standardize the sparsity file format
        if len(sparsity) > 0 and sparsity[0] >= 0:
            sparsity = torch.log10(sparsity + 1e-10)
        sparsity = sparsity.to(self.device)
        self.target_feature_indexes = (
            (sparsity > self.sparsity_threshold).nonzero(as_tuple=True)[0].tolist()
        )

        # divide into batches
        feature_idx = torch.tensor(self.target_feature_indexes)
        n_subarrays = np.ceil(len(feature_idx) / self.n_features_at_a_time).astype(int)
        feature_idx = np.array_split(feature_idx, n_subarrays)
        feature_idx = [x.tolist() for x in feature_idx]

        # print(f"==== Starting Batch: {self.start_batch}")
        # if self.end_batch is not None and self.end_batch != self.start_batch:
        #     print(f"==== Ending at Batch: {self.end_batch}")

        if self.start_batch > len(feature_idx) + 1:
            print(
                f"Start batch {self.start_batch} is greater than number of batches + 1 {len(feature_idx)}, exiting"
            )
            exit()

        # write dead into file so we can create them as dead in Neuronpedia
        skipped_indexes = set(range(self.n_features)) - set(self.target_feature_indexes)
        skipped_indexes_json = json.dumps(
            {
                "model_id": self.model_id,
                "layer": str(self.layer),
                "sae_id": self.sae_id,
                "log_sparsity": self.sparsity_threshold,
                "skipped_indexes": list(skipped_indexes),
            }
        )
        with open(f"{self.outputs_dir}/skipped_indexes.json", "w") as f:
            f.write(skipped_indexes_json)

        tokens_file = f"{self.outputs_dir}/tokens_{self.n_batches_to_sample_from}_{self.n_prompts_to_select}.pt"
        if os.path.isfile(tokens_file):
            print("Tokens exist, loading them.")
            tokens = torch.load(tokens_file)
        else:
            print("Tokens don't exist, making them.")
            tokens = self.get_tokens(
                self.n_batches_to_sample_from, self.n_prompts_to_select
            )
            torch.save(
                tokens,
                tokens_file,
            )

        vocab_dict = self.model.tokenizer.vocab
        new_vocab_dict = {}
        # Replace substrings in the keys of vocab_dict using HTML_ANOMALIES
        for k, v in vocab_dict.items():
            modified_key = k
            for anomaly in HTML_ANOMALIES:
                modified_key = modified_key.replace(anomaly, HTML_ANOMALIES[anomaly])
            new_vocab_dict[v] = modified_key
        vocab_dict = new_vocab_dict

        # pad with blank tokens to the actual vocab size
        for i in range(len(vocab_dict), self.model.cfg.d_vocab):
            vocab_dict[i] = OUT_OF_RANGE_TOKEN

        with torch.no_grad():
            feature_batch_count = 0
            for features_to_process in tqdm(feature_idx):
                feature_batch_count = feature_batch_count + 1

                if feature_batch_count < self.start_batch:
                    # print(f"Skipping batch - it's after start_batch: {feature_batch_count}")
                    continue
                if self.end_batch is not None and feature_batch_count > self.end_batch:
                    # print(f"Skipping batch - it's after end_batch: {feature_batch_count}")
                    continue

                output_file = f"{self.outputs_dir}/batch-{feature_batch_count}.json"
                # if output_file exists, skip
                if os.path.isfile(output_file):
                    logline = f"\n++++++++++ Skipping Batch #{feature_batch_count} output. File exists: {output_file} ++++++++++\n"
                    print(logline)
                    continue

                print(f"========== Running Batch #{feature_batch_count} ==========")

                layout = SaeVisLayoutConfig(
                    columns=[
                        Column(
                            SequencesConfig(
                                stack_mode="stack-all",
                                buffer=None,  # type: ignore
                                compute_buffer=True,
                                n_quantiles=self.n_quantiles,
                                top_acts_group_size=self.top_acts_group_size,
                                quantile_group_size=self.quantile_group_size,
                            ),
                            ActsHistogramConfig(),
                            LogitsHistogramConfig(),
                            LogitsTableConfig(),
                            FeatureTablesConfig(n_rows=3),
                        )
                    ]
                )
                feature_vis_params = SaeVisConfig(
                    hook_point=self.sparse_autoencoder.cfg.hook_point,
                    minibatch_size_features=128,
                    minibatch_size_tokens=64,
                    features=features_to_process,
                    verbose=True,
                    feature_centric_layout=layout,
                )
                feature_data = SaeVisData.create(
                    encoder=self.sparse_autoencoder,  # type: ignore
                    model=cast(HookedTransformer, self.model),
                    tokens=tokens,
                    cfg=feature_vis_params,
                )

                features_outputs = []
                for _, feat_index in enumerate(feature_data.feature_data_dict.keys()):
                    feature = feature_data.feature_data_dict[feat_index]

                    feature_output = {}
                    feature_output["featureIndex"] = feat_index

                    top10_logits = self.round_list(feature.logits_table_data.top_logits)
                    bottom10_logits = self.round_list(
                        feature.logits_table_data.bottom_logits
                    )

                    if feature.feature_tables_data:
                        feature_output["neuron_alignment_indices"] = (
                            feature.feature_tables_data.neuron_alignment_indices
                        )
                        feature_output["neuron_alignment_values"] = self.round_list(
                            feature.feature_tables_data.neuron_alignment_values
                        )
                        feature_output["neuron_alignment_l1"] = self.round_list(
                            feature.feature_tables_data.neuron_alignment_l1
                        )
                        feature_output["correlated_neurons_indices"] = (
                            feature.feature_tables_data.correlated_neurons_indices
                        )
                        feature_output["correlated_neurons_l1"] = self.round_list(
                            feature.feature_tables_data.correlated_neurons_cossim
                        )
                        feature_output["correlated_neurons_pearson"] = self.round_list(
                            feature.feature_tables_data.correlated_neurons_pearson
                        )
                        feature_output["correlated_features_indices"] = (
                            feature.feature_tables_data.correlated_features_indices
                        )
                        feature_output["correlated_features_l1"] = self.round_list(
                            feature.feature_tables_data.correlated_features_cossim
                        )
                        feature_output["correlated_features_pearson"] = self.round_list(
                            feature.feature_tables_data.correlated_features_pearson
                        )

                    feature_output["neg_str"] = self.to_str_tokens_safe(
                        vocab_dict, feature.logits_table_data.bottom_token_ids
                    )
                    feature_output["neg_values"] = bottom10_logits
                    feature_output["pos_str"] = self.to_str_tokens_safe(
                        vocab_dict, feature.logits_table_data.top_token_ids
                    )
                    feature_output["pos_values"] = top10_logits

                    feature_output["frac_nonzero"] = (
                        float(
                            feature.acts_histogram_data.title.split(" = ")[1].split(
                                "%"
                            )[0]
                        )
                        / 100
                        if feature.acts_histogram_data.title is not None
                        else 0
                    )

                    freq_hist_data = feature.acts_histogram_data
                    freq_bar_values = self.round_list(freq_hist_data.bar_values)
                    feature_output["freq_hist_data_bar_values"] = freq_bar_values
                    feature_output["freq_hist_data_bar_heights"] = self.round_list(
                        freq_hist_data.bar_heights
                    )

                    logits_hist_data = feature.logits_histogram_data
                    feature_output["logits_hist_data_bar_heights"] = self.round_list(
                        logits_hist_data.bar_heights
                    )
                    feature_output["logits_hist_data_bar_values"] = self.round_list(
                        logits_hist_data.bar_values
                    )

                    feature_output["num_tokens_for_dashboard"] = (
                        self.n_prompts_to_select
                    )

                    activations = []
                    sdbs = feature.sequence_data
                    for sgd in sdbs.seq_group_data:
                        binMin = 0
                        binMax = 0
                        binContains = 0
                        if "TOP ACTIVATIONS" in sgd.title:
                            binMin = -1
                            try:
                                binMax = float(sgd.title.split(" = ")[-1])
                            except ValueError:
                                print(f"Error parsing top acts: {sgd.title}")
                                binMax = 99
                            binContains = -1
                        elif "INTERVAL" in sgd.title:
                            try:
                                split = sgd.title.split("<br>")
                                firstSplit = split[0].split(" ")
                                binMin = float(firstSplit[1])
                                binMax = float(firstSplit[-1])
                                secondSplit = split[1].split(" ")
                                binContains = float(secondSplit[-1].rstrip("%")) / 100
                            except ValueError:
                                print(f"Error parsing interval: {sgd.title}")
                        for sd in sgd.seq_data:
                            if (
                                sd.top_token_ids is not None
                                and sd.bottom_token_ids is not None
                                and sd.top_logits is not None
                                and sd.bottom_logits is not None
                            ):
                                activation = {}
                                activation["binMin"] = binMin
                                activation["binMax"] = binMax
                                activation["binContains"] = binContains

                                strs = []
                                posContribs = []
                                negContribs = []
                                for i in range(len(sd.token_ids)):
                                    strs.append(
                                        self.to_str_tokens_safe(
                                            vocab_dict, sd.token_ids[i]
                                        )
                                    )
                                    posContrib = {}
                                    posTokens = [
                                        self.to_str_tokens_safe(vocab_dict, j)
                                        for j in sd.top_token_ids[i]
                                    ]
                                    if len(posTokens) > 0:
                                        posContrib["t"] = posTokens
                                        posContrib["v"] = self.round_list(
                                            sd.top_logits[i]
                                        )
                                    posContribs.append(posContrib)
                                    negContrib = {}
                                    negTokens = [
                                        self.to_str_tokens_safe(vocab_dict, j)  # type: ignore
                                        for j in sd.bottom_token_ids[i]
                                    ]
                                    if len(negTokens) > 0:
                                        negContrib["t"] = negTokens
                                        negContrib["v"] = self.round_list(
                                            sd.bottom_logits[i]
                                        )
                                    negContribs.append(negContrib)

                                activation["logitContributions"] = json.dumps(
                                    {"pos": posContribs, "neg": negContribs}
                                )
                                activation["tokens"] = strs
                                activation["values"] = self.round_list(sd.feat_acts)
                                activation["maxValue"] = max(activation["values"])
                                activation["lossValues"] = self.round_list(
                                    sd.loss_contribution
                                )

                                activations.append(activation)
                    feature_output["activations"] = activations

                    features_outputs.append(feature_output)

                to_write = {
                    "model_id": self.model_id,
                    "layer": str(self.layer),
                    "sae_id": self.sae_id,
                    "features": features_outputs,
                    "n_batches_to_sample_from": self.n_batches_to_sample_from,
                    "n_prompts_to_select": self.n_prompts_to_select,
                }
                json_object = json.dumps(to_write, cls=NpEncoder)

                with open(
                    output_file,
                    "w",
                ) as f:
                    f.write(json_object)

                logline = f"\n========== Completed Batch #{feature_batch_count} output: {output_file} ==========\n"

        return



import traceback
from typing import Any, cast

import wandb

from sae_lens.training.config import LanguageModelSAERunnerConfig
from sae_lens.training.load_model import load_model
from sae_lens.training.session_loader import LMSparseAutoencoderSessionloader
from sae_lens.training.train_sae_on_language_model import (
    load_checkpoint,
    train_sae_group_on_language_model,
)


def language_model_sae_runner(cfg: LanguageModelSAERunnerConfig):
    """ """
    training_run_state = None
    train_contexts = None

    if cfg.resume:
        try:
            checkpoint_path = cfg.get_resume_checkpoint_path()
            model = load_model(
                model_class_name=cfg.model_class_name,
                model_name=cfg.model_name,
                device=cfg.device,
            )
            model.to(cfg.device)
            (
                training_run_state,
                activations_loader,
                sparse_autoencoder,
                train_contexts,
            ) = load_checkpoint(
                checkpoint_path=checkpoint_path,
                cfg=cfg,
                model=model,
                batch_size=cfg.train_batch_size,
            )
        # no checkpoints found, don't resume
        except FileNotFoundError:
            print(traceback.format_exc())
            print("failed to find checkpoint to resume from, setting resume to False")
            cfg.resume = False

    if not cfg.resume:
        if cfg.from_pretrained_path is not None:
            (
                model,
                sparse_autoencoder,
                activations_loader,
            ) = LMSparseAutoencoderSessionloader.load_pretrained_sae(
                cfg.from_pretrained_path
            )
            cfg = sparse_autoencoder.cfg
        else:
            loader = LMSparseAutoencoderSessionloader(cfg)
            model, sparse_autoencoder, activations_loader = (
                loader.load_sae_training_group_session()
            )

    if cfg.log_to_wandb:
        resume = None
        if cfg.resume:
            resume = "allow"
        wandb.init(
            project=cfg.wandb_project,
            config=cast(Any, cfg),
            name=cfg.run_name,
            resume=resume,
            id=cfg.wandb_id,
        )

    # train SAE
    sparse_autoencoder = train_sae_group_on_language_model(
        model=model,  # pyright: ignore [reportPossiblyUnboundVariable]
        sae_group=sparse_autoencoder,  # pyright: ignore [reportPossiblyUnboundVariable]
        activation_store=activations_loader,  # pyright: ignore [reportPossiblyUnboundVariable]
        train_contexts=train_contexts,
        training_run_state=training_run_state,
        batch_size=cfg.train_batch_size,
        n_checkpoints=cfg.n_checkpoints,
        feature_sampling_window=cfg.feature_sampling_window,
        use_wandb=cfg.log_to_wandb,
        wandb_log_frequency=cfg.wandb_log_frequency,
    ).sae_group

    if cfg.log_to_wandb:
        wandb.finish()

    return sparse_autoencoder



import os
from typing import Any, Optional, cast

# set TOKENIZERS_PARALLELISM to false to avoid warnings
os.environ["TOKENIZERS_PARALLELISM"] = "false"
import shutil
import time
import uuid

import pandas as pd
import plotly
import plotly.express as px
import torch
import wandb
from sae_vis.data_config_classes import (
    ActsHistogramConfig,
    Column,
    FeatureTablesConfig,
    SaeVisConfig,
    SaeVisLayoutConfig,
    SequencesConfig,
)
from sae_vis.data_fetching_fns import get_feature_data
from torch.nn.functional import cosine_similarity
from tqdm import tqdm
from transformer_lens import HookedTransformer

from sae_lens.training.session_loader import LMSparseAutoencoderSessionloader


class DashboardRunner:

    model: HookedTransformer | None = None

    def __init__(
        self,
        sae_path: Optional[str] = None,
        dashboard_parent_folder: str = "./feature_dashboards",
        wandb_artifact_path: Optional[str] = None,
        init_session: bool = True,
        # token pars
        n_batches_to_sample_from: int = 2**12,
        n_prompts_to_select: int = 4096 * 6,
        # sampling pars
        n_features_at_a_time: int = 1024,
        max_batch_size: int = 256,
        buffer_tokens: int = 8,
        # util pars
        use_wandb: bool = False,
        continue_existing_dashboard: bool = True,
        final_index: Optional[int] = None,
    ):
        """ """

        if wandb_artifact_path is not None:
            artifact_dir = f"artifacts/{wandb_artifact_path.split('/')[2]}"
            if not os.path.exists(artifact_dir):
                print("Downloading artifact")
                run = wandb.init()
                assert run is not None  # keep pyright happy
                artifact = run.use_artifact(wandb_artifact_path)
                artifact_dir = artifact.download()
                path_to_artifact = f"{artifact_dir}/{os.listdir(artifact_dir)[0]}"
                # feature sparsity
                feature_sparsity_path = self.get_feature_sparsity_path(
                    wandb_artifact_path
                )
                artifact = run.use_artifact(feature_sparsity_path)
                artifact_dir = artifact.download()
                # add it as a property
                self.feature_sparsity = torch.load(
                    f"{artifact_dir}/{os.listdir(artifact_dir)[0]}"
                )
            else:
                print("Artifact already downloaded")
                path_to_artifact = f"{artifact_dir}/{os.listdir(artifact_dir)[0]}"

                feature_sparsity_path = self.get_feature_sparsity_path(
                    wandb_artifact_path
                )
                artifact_dir = f"artifacts/{feature_sparsity_path.split('/')[2]}"
                feature_sparsity_file = os.listdir(artifact_dir)[0]
                self.feature_sparsity = torch.load(
                    f"{artifact_dir}/{feature_sparsity_file}"
                )

            self.sae_path = path_to_artifact
        else:
            assert sae_path is not None
            self.sae_path = sae_path
            self.feature_sparsity = None

        if init_session:
            self.init_sae_session()

        self.n_features_at_a_time = n_features_at_a_time
        self.max_batch_size = max_batch_size
        self.buffer_tokens = buffer_tokens
        self.use_wandb = use_wandb
        self.final_index = (
            final_index
            if final_index is not None
            else self.sparse_autoencoder.cfg.d_sae
        )
        self.n_batches_to_sample_from = n_batches_to_sample_from
        self.n_prompts_to_select = n_prompts_to_select

        # Deal with file structure
        if not os.path.exists(dashboard_parent_folder):
            os.makedirs(dashboard_parent_folder)

        self.dashboard_folder = (
            f"{dashboard_parent_folder}/{self.get_dashboard_folder_name()}"
        )
        if not os.path.exists(self.dashboard_folder):
            os.makedirs(self.dashboard_folder)

        if not continue_existing_dashboard:
            # check if there are files there and if so abort
            if len(os.listdir(self.dashboard_folder)) > 0:
                raise ValueError("Dashboard folder not empty. Aborting.")

    def get_feature_sparsity_path(self, wandb_artifact_path: str):
        prefix = wandb_artifact_path.split(":")[0]
        return f"{prefix}_log_feature_sparsity:v9"

    def get_dashboard_folder_name(self):
        model = self.sparse_autoencoder.cfg.model_name
        hook_point = self.sparse_autoencoder.cfg.hook_point
        d_sae = self.sparse_autoencoder.cfg.d_sae
        dashboard_folder_name = f"{model}_{hook_point}_{d_sae}"

        return dashboard_folder_name

    def init_sae_session(self):
        (
            model,
            sae_group,
            self.activation_store,
        ) = LMSparseAutoencoderSessionloader.load_pretrained_sae(self.sae_path)
        assert isinstance(
            model, HookedTransformer
        )  # only HookedTransformer is allowed to be used in the dashboard
        self.model = model
        # TODO: handle multiple autoencoders
        self.sparse_autoencoder = next(iter(sae_group))[1]

    def get_tokens(
        self, n_batches_to_sample_from: int = 2**12, n_prompts_to_select: int = 4096 * 6
    ):
        """
        Get the tokens needed for dashboard generation.
        """

        all_tokens_list = []
        pbar = tqdm(range(n_batches_to_sample_from))
        for _ in pbar:
            batch_tokens = self.activation_store.get_batch_tokens()
            batch_tokens = batch_tokens[torch.randperm(batch_tokens.shape[0])][
                : batch_tokens.shape[0]
            ]
            all_tokens_list.append(batch_tokens)

        all_tokens = torch.cat(all_tokens_list, dim=0)
        all_tokens = all_tokens[torch.randperm(all_tokens.shape[0])]
        return all_tokens[:n_prompts_to_select]

    def get_index_to_resume_from(self):
        i = 0
        assert self.n_features is not None  # keep pyright happy
        for i in range(self.n_features):
            if not os.path.exists(f"{self.dashboard_folder}/data_{i:04}.html"):
                break

        assert self.sparse_autoencoder.cfg.d_sae is not None  # keep pyright happy
        assert self.final_index is not None  # keep pyright happy
        n_features = self.sparse_autoencoder.cfg.d_sae
        n_features_at_a_time = self.n_features_at_a_time
        id_of_last_feature_without_dashboard = i
        n_features_remaining = self.final_index - id_of_last_feature_without_dashboard
        n_batches_to_do = n_features_remaining // n_features_at_a_time
        if self.final_index == n_features:
            id_to_start_from = max(
                0, n_features - (n_batches_to_do + 1) * n_features_at_a_time
            )
        else:
            id_to_start_from = 0  # testing purposes only

        print(f"File {i} does not exist")
        print(f"features left to do: {n_features_remaining}")
        print(f"id_to_start_from: {id_to_start_from}")
        print(
            f"number of batches to do: {(n_features - id_to_start_from) // n_features_at_a_time}"
        )

        return id_to_start_from

    @torch.no_grad()
    def get_feature_property_df(self):
        sparse_autoencoder = self.sparse_autoencoder
        feature_sparsity = (
            self.feature_sparsity
            if self.feature_sparsity is not None
            else torch.tensor(0)
        )

        W_dec_normalized = (
            sparse_autoencoder.W_dec.cpu()
        )  # / sparse_autoencoder.W_dec.cpu().norm(dim=-1, keepdim=True)
        W_enc_normalized = (
            sparse_autoencoder.W_enc.cpu()
            / sparse_autoencoder.W_enc.cpu().norm(dim=-1, keepdim=True)
        )
        d_e_projection = cosine_similarity(W_dec_normalized, W_enc_normalized.T)

        assert sparse_autoencoder.cfg.d_sae is not None  # keep pyright happy
        temp_df = pd.DataFrame(
            {
                "log_feature_sparsity": feature_sparsity + 1e-10,
                "d_e_projection": d_e_projection,
                # "d_e_projection_normalized": d_e_projection_normalized,
                "b_enc": sparse_autoencoder.b_enc.detach().cpu(),
                "feature": [
                    f"feature_{i}" for i in range(sparse_autoencoder.cfg.d_sae)
                ],
                "index": torch.arange(sparse_autoencoder.cfg.d_sae),
                "dead_neuron": (feature_sparsity < -9).cpu(),
            }
        )

        return temp_df

    def run(self):
        """
        Generate the dashboard.
        """

        run = None
        if self.use_wandb:
            # get name from wandb
            random_suffix = str(uuid.uuid4())[:8]
            name = f"{self.get_dashboard_folder_name()}_{random_suffix}"
            run = wandb.init(
                project="feature_dashboards",
                config=cast(Any, self.sparse_autoencoder.cfg),
                name=name,
                tags=[
                    f"model_{self.sparse_autoencoder.cfg.model_name}",
                    f"hook_point_{self.sparse_autoencoder.cfg.hook_point}",
                ],
            )

        if self.model is None:
            self.init_sae_session()

        # generate all the plots
        if self.use_wandb and self.feature_sparsity is not None:
            feature_property_df = self.get_feature_property_df()

            fig = px.histogram(
                self.feature_sparsity + 1e-10,
                nbins=100,
                log_x=False,
                title="Feature sparsity",
            )
            wandb.log(
                {"plots/feature_density_histogram": wandb.Html(plotly.io.to_html(fig))}
            )

            fig = px.histogram(
                self.sparse_autoencoder.b_enc.detach().cpu(), title="b_enc", nbins=100
            )
            wandb.log({"plots/b_enc_histogram": wandb.Html(plotly.io.to_html(fig))})

            fig = px.histogram(
                feature_property_df.d_e_projection, nbins=100, title="D/E projection"
            )
            wandb.log(
                {"plots/d_e_projection_histogram": wandb.Html(plotly.io.to_html(fig))}
            )

            fig = px.histogram(
                self.sparse_autoencoder.b_dec.detach().cpu(),
                nbins=100,
                title="b_dec projection onto W_dec",
            )
            wandb.log(
                {"plots/b_dec_projection_histogram": wandb.Html(plotly.io.to_html(fig))}
            )

            fig = px.scatter_matrix(
                feature_property_df,
                dimensions=["log_feature_sparsity", "d_e_projection", "b_enc"],
                color="dead_neuron",
                hover_name="feature",
                opacity=0.2,
                height=800,
                width=1400,
            )
            wandb.log({"plots/scatter_matrix": wandb.Html(plotly.io.to_html(fig))})

        self.n_features = self.sparse_autoencoder.cfg.d_sae
        id_to_start_from = self.get_index_to_resume_from()
        id_to_end_at = self.n_features if self.final_index is None else self.final_index
        assert id_to_end_at is not None  # keep pyright happy

        # divide into batches
        feature_idx = torch.tensor(range(id_to_start_from, id_to_end_at))
        feature_idx = feature_idx.reshape(-1, self.n_features_at_a_time)
        feature_idx = [x.tolist() for x in feature_idx]

        print(f"Hook Point Layer: {self.sparse_autoencoder.cfg.hook_point_layer}")
        print(f"Hook Point: {self.sparse_autoencoder.cfg.hook_point}")
        print(f"Writing files to: {self.dashboard_folder}")

        # get tokens:
        start = time.time()
        tokens = self.get_tokens(
            self.n_batches_to_sample_from, self.n_prompts_to_select
        )
        end = time.time()
        print(f"Time to get tokens: {end - start}")
        if self.use_wandb:
            wandb.log({"time/time_to_get_tokens": end - start})

        assert self.model is not None
        vocab_dict = cast(Any, self.model.tokenizer).vocab
        vocab_dict = {
            v: k.replace("Ġ", " ").replace("\n", "\\n") for k, v in vocab_dict.items()
        }

        with torch.no_grad():
            for interesting_features in tqdm(feature_idx):
                print(interesting_features)

                layout = SaeVisLayoutConfig(
                    columns=[
                        Column(
                            SequencesConfig(
                                stack_mode="stack-all",
                                buffer=(self.buffer_tokens, self.buffer_tokens),
                                compute_buffer=False,
                                n_quantiles=10,
                                top_acts_group_size=20,
                                quantile_group_size=5,
                            ),
                            width=650,
                        ),
                        Column(
                            ActsHistogramConfig(),
                            FeatureTablesConfig(n_rows=5),
                            width=500,
                        ),
                    ],
                    height=1000,
                )
                feature_vis_params = SaeVisConfig(
                    hook_point=self.sparse_autoencoder.cfg.hook_point,
                    minibatch_size_features=256,
                    minibatch_size_tokens=64,
                    features=interesting_features,
                    verbose=True,
                    feature_centric_layout=layout,
                )

                feature_data = get_feature_data(
                    encoder=self.sparse_autoencoder,  # type: ignore
                    model=self.model,
                    tokens=tokens,
                    cfg=feature_vis_params,
                )

                for i, test_idx in enumerate(feature_data.feature_data_dict.keys()):
                    feature_data.save_feature_centric_vis(
                        f"{self.dashboard_folder}/data_{test_idx:04}.html",
                        feature_idx=test_idx,
                    )

                    if i < 10 and self.use_wandb:
                        # upload the html as an artifact
                        artifact = wandb.Artifact(f"feature_{test_idx}", type="feature")
                        artifact.add_file(
                            f"{self.dashboard_folder}/data_{test_idx:04}.html"
                        )
                        assert run is not None  # keep pyright happy
                        run.log_artifact(artifact)

                        # also upload as html to dashboard
                        wandb.log(
                            {
                                "features/feature_dashboard": wandb.Html(
                                    f"{self.dashboard_folder}/data_{test_idx:04}.html"
                                )
                            },
                            step=test_idx,
                        )

        if self.use_wandb:
            # when done zip the folder
            shutil.make_archive(self.dashboard_folder, "zip", self.dashboard_folder)

            # then upload the zip as an artifact
            artifact = wandb.Artifact("dashboard", type="zipped_feature_dashboards")
            artifact.add_file(f"{self.dashboard_folder}.zip")
            assert run is not None  # keep pyright happy
            run.log_artifact(artifact)

            # terminate the run
            run.finish()

            # delete the dashboard folder
            shutil.rmtree(self.dashboard_folder)

        return



from typing import Any, cast

import torch
from transformer_lens import HookedTransformer
from transformer_lens.hook_points import HookedRootModule


def load_model(
    model_class_name: str, model_name: str, device: str | torch.device | None = None
) -> HookedRootModule:
    if model_class_name == "HookedTransformer":
        return HookedTransformer.from_pretrained(model_name=model_name, device=device)
    elif model_class_name == "HookedMamba":
        try:
            from mamba_lens import HookedMamba
        except ImportError:
            raise ValueError(
                "mamba-lens must be installed to work with mamba models. This can be added with `pip install sae-lens[mamba]`"
            )
        # HookedMamba has incorrect typing information, so we need to cast the type here
        return cast(
            HookedRootModule,
            HookedMamba.from_pretrained(model_name, device=cast(Any, device)),
        )
    else:
        raise ValueError(f"Unknown model class: {model_class_name}")



import json
import os
import pathlib
from typing import Optional, Tuple

import torch
from huggingface_hub import hf_hub_download, list_repo_tree
from safetensors import safe_open
from tqdm import tqdm

from sae_lens.training.config import LanguageModelSAERunnerConfig
from sae_lens.training.sparse_autoencoder import SparseAutoencoder


def load_sparsity(path: str) -> torch.Tensor:
    sparsity_path = os.path.join(path, "sparsity.safetensors")
    with safe_open(sparsity_path, framework="pt", device="cpu") as f:  # type: ignore
        sparsity = f.get_tensor("sparsity")
    return sparsity


def download_sae_from_hf(
    repo_id: str = "jbloom/GPT2-Small-SAEs-Reformatted",
    folder_name: str = "blocks.0.hook_resid_pre",
    force_download: bool = False,
) -> Tuple[str, str, Optional[str]]:

    FILENAME = f"{folder_name}/cfg.json"
    cfg_path = hf_hub_download(
        repo_id=repo_id, filename=FILENAME, force_download=force_download
    )

    FILENAME = f"{folder_name}/sae_weights.safetensors"
    sae_path = hf_hub_download(
        repo_id=repo_id, filename=FILENAME, force_download=force_download
    )

    try:
        FILENAME = f"{folder_name}/sparsity.safetensors"
        sparsity_path = hf_hub_download(
            repo_id=repo_id, filename=FILENAME, force_download=force_download
        )
    except:  # noqa
        sparsity_path = None

    return cfg_path, sae_path, sparsity_path


def load_sae_from_local_path(path: str) -> Tuple[SparseAutoencoder, torch.Tensor]:
    sae = SparseAutoencoder.load_from_pretrained(path)
    sparsity = load_sparsity(path)
    return sae, sparsity


def get_gpt2_res_jb_saes(
    hook_point: Optional[str] = None,
    device: str = "cpu",
) -> tuple[dict[str, SparseAutoencoder], dict[str, torch.Tensor]]:
    """
    Download the sparse autoencoders for the GPT2-Small model with residual connections
    from the repository of jbloom. You can specify a hook_point to download only one
    of the sparse autoencoders if desired.

    """

    GPT2_SMALL_RESIDUAL_SAES_REPO_ID = "jbloom/GPT2-Small-SAEs-Reformatted"
    GPT2_SMALL_RESIDUAL_SAES_HOOK_POINTS = [
        f"blocks.{layer}.hook_resid_pre" for layer in range(12)
    ] + ["blocks.11.hook_resid_post"]

    if hook_point is not None:
        assert hook_point in GPT2_SMALL_RESIDUAL_SAES_HOOK_POINTS, (
            f"hook_point must be one of {GPT2_SMALL_RESIDUAL_SAES_HOOK_POINTS}"
            f"but got {hook_point}"
        )
        GPT2_SMALL_RESIDUAL_SAES_HOOK_POINTS = [hook_point]

    saes = {}
    sparsities = {}
    for hook_point in tqdm(GPT2_SMALL_RESIDUAL_SAES_HOOK_POINTS):

        _, sae_path, _ = download_sae_from_hf(
            repo_id=GPT2_SMALL_RESIDUAL_SAES_REPO_ID, folder_name=hook_point
        )

        # Then use our function to download the files
        folder_path = os.path.dirname(sae_path)
        sae = SparseAutoencoder.load_from_pretrained(folder_path, device=device)
        sparsity = load_sparsity(folder_path)
        sparsity = sparsity.to(device)
        saes[hook_point] = sae
        sparsities[hook_point] = sparsity

    return saes, sparsities


def convert_connor_rob_sae_to_our_saelens_format(
    state_dict: dict[str, torch.Tensor],
    config: dict[str, int | str],
    device: str = "cpu",
):
    """

    # can get session like so.
        model, ae_alt, activation_store = LMSparseAutoencoderSessionloader(
            cfg
        ).load_sae_training_group_session()
        next(iter(ae_alt))[1].load_state_dict(state_dict)
        return model, ae_alt, activation_store

    """

    expansion_factor = int(config["dict_size"]) // int(config["act_size"])

    cfg = LanguageModelSAERunnerConfig(
        model_name=config["model_name"],  # type: ignore
        hook_point=config["act_name"],  # type: ignore
        hook_point_layer=config["layer"],  # type: ignore
        # data
        # dataset_path = "/share/data/datasets/pile/the-eye.eu/public/AI/pile/train", # Training set of The Pile
        dataset_path="NeelNanda/openwebtext-tokenized-9b",
        is_dataset_tokenized=True,
        d_in=config["act_size"],  # type: ignore
        expansion_factor=expansion_factor,
        context_size=config["seq_len"],  # type: ignore
        device=device,
        store_batch_size=32,
        n_batches_in_buffer=10,
        prepend_bos=False,
        verbose=False,
        dtype=torch.float32,
    )

    ae_alt = SparseAutoencoder(cfg)
    ae_alt.load_state_dict(state_dict)
    return ae_alt


def convert_old_to_modern_saelens_format(
    pytorch_file: str, out_folder: str = "", force: bool = False
):
    """
    Reads a pretrained SAE from the old pickle-style SAELens .pt format, then saves a modern-format SAELens SAE.

    Arguments:
    ----------
    pytorch_file: str
        Path of old format file to open.
    out_folder: str, optional
        Path where new SAE will be stored; if None, out_folder = pytorch_file with the '.pt' removed.
    force: bool, optional
        If out_folder already exists, this function will not save unless force=True.
    """
    file_path = pathlib.Path(pytorch_file)
    if out_folder == "":
        out_f = file_path.parent / file_path.stem
    else:
        out_f = pathlib.Path(out_folder)
    if (not force) and out_f.exists():
        raise FileExistsError(f"{out_folder} already exists and force=False")
    out_f.mkdir(exist_ok=True, parents=True)

    # Load model & save in new format.
    autoencoder = SparseAutoencoder.load_from_pretrained_legacy(str(file_path))
    autoencoder.save_model(str(out_f))


def get_gpt2_small_ckrk_attn_out_saes() -> dict[str, SparseAutoencoder]:

    REPO_ID = "ckkissane/attn-saes-gpt2-small-all-layers"

    # list all files in repo
    saes_weights = {}
    sae_configs = {}
    repo_files = list_repo_tree(REPO_ID)
    for i in tqdm(repo_files):
        file_name = i.path
        if file_name.endswith(".pt"):
            # print(f"Downloading {file_name}")
            path = hf_hub_download(REPO_ID, file_name)
            name = path.split("/")[-1].split(".pt")[0]
            saes_weights[name] = torch.load(path, map_location="mps")
        elif file_name.endswith(".json"):
            # print(f"Downloading {file_name}")
            config_path = hf_hub_download(REPO_ID, file_name)
            name = config_path.split("/")[-1].split("_cfg.json")[0]
            sae_configs[name] = json.load(open(config_path, "r"))

    saes = {}
    for name, config in sae_configs.items():
        print(f"Loading {name}")
        saes[name] = convert_connor_rob_sae_to_our_saelens_format(
            saes_weights[name], config
        )

    return saes
